"""Handles reading and writing the WRF-Ensembly Observation data files."""

from pathlib import Path

import numpy as np
import pandas as pd

REQUIRED_COLUMNS = [
    "instrument",
    "quantity",
    "time",
    "longitude",
    "latitude",
    "z",
    "z_type",
    "value",
    "value_uncertainty",
    "qc_flag",
    "orig_coords",
    "orig_filename",
    "metadata",
]

Z_TYPES = ["surface", "pressure", "height", "model_level", "columnar"]


def validate_schema(df: pd.DataFrame):
    """
    Checks the following for the input dataframe:
    - All required columns are present
    - z_type values are valid
    - orig_coords is a dictionary with keys 'indices', 'shape', and 'names'
    - orig_coords 'indices', 'shape', and 'names' have the same length
    - orig_coords 'indices' are integers
    - orig_coords 'shape' are integers
    - orig_coords 'names' are strings
    Throws a ValueError if any checks fail.

    Args:
        df: The dataframe to check
    """

    missing = set(REQUIRED_COLUMNS) - set(df.columns)
    if missing:
        raise ValueError(f"Missing columns: {missing}")

    # Verify z_type values, orig_coords fields
    for i, row in df.iterrows():
        if row.z_type not in Z_TYPES:
            raise ValueError(f"Invalid z_type encountered at row {i}: {row.z_type}")

        orig_coords = row.orig_coords
        if not isinstance(orig_coords, dict):
            raise ValueError(f"orig_coords must be a dictionary at row {i}")

        if (
            "indices" not in orig_coords
            or "shape" not in orig_coords
            or "names" not in orig_coords
        ):
            raise ValueError(
                f"orig_coords must contain 'indices', 'shape', and 'names' keys at row {i}"
            )

        if (
            len(orig_coords["indices"])
            != len(orig_coords["names"])
            != len(orig_coords["shape"])
        ):
            raise ValueError(
                f"orig_coords 'indices', 'shape', and 'names' must have the same length at row {i}"
            )
        orig_coord_len = len(orig_coords["indices"])
        for j in range(orig_coord_len):
            if not isinstance(orig_coords["indices"][j], (int, np.integer)):
                print(type(orig_coords["indices"][j]))
                raise ValueError(f"orig_coords 'indices' must be integers at row {i}")
            if not isinstance(orig_coords["shape"][j], (int, np.integer)):
                raise ValueError(f"orig_coords 'shape' must be integers at row {i}")
            if not isinstance(orig_coords["names"][j], (str, np.str_)):
                raise ValueError(f"orig_coords 'names' must be strings at row {i}")


def read_obs(path: Path | str) -> pd.DataFrame:
    """Read a WRF-Ensembly Observation data file into a pandas DataFrame."""

    df = pd.read_parquet(path)
    validate_schema(df)
    return df


def write_obs(df: pd.DataFrame, path: Path | str):
    """Write a pandas DataFrame to a WRF-Ensembly Observation data file."""

    validate_schema(df)
    df.to_parquet(path, index=False)



import datetime as dt
import re
from itertools import chain
from pathlib import Path

import click
import tomli_w


@click.command()
@click.argument("dir_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(path_type=Path))
def create_aeolus_l2a_obsgroup(dir_path: Path, output_path: Path):
    """
    Creates the observation group file for the given directory of AEOLUS L2B DBL files
    """

    prefix = "AE_OPER_ALD_U_N_2A_"
    date_regex = re.compile(r"(\d{8}T\d{9})_(\d{9})")

    files = []
    for dbl in chain(dir_path.glob("*.dbl"), dir_path.glob("*.DBL")):
        click.echo(f"Processing {dbl}")

        # Sample file name: AE_OPER_ALD_U_N_2A_20210903T164708024_005532006_017567_0001.DBL
        # From docs: yyyymmddThhmmsszzz_uuuuuuuuu_oooooo_vvvv
        # yyyymmddThhmmsszzz: start time of the measurement
        # uuuuuuuuu: duration of the measurement in milliseconds
        # oooooo: orbit number
        # vvvv: version number
        name = dbl.name
        if name[: len(prefix)] != prefix:
            click.echo(f"File prefix is not {prefix}!")
            continue

        start, end = date_regex.findall(name)[0]
        start = dt.datetime.strptime(start[:-3], "%Y%m%dT%H%M%S").replace(
            tzinfo=dt.timezone.utc
        )
        end = start + dt.timedelta(milliseconds=int(end))

        files.append({"start_date": start, "end_date": end, "path": str(dbl.resolve())})

    if not files:
        click.echo("No files found!")
        return

    # Sort files by start date, easier to browse the file this way
    files = sorted(files, key=lambda f: f["start_date"])

    obsgroup = {
        "kind": "AEOLUS_L2A",
        "converter": "/path/to/convert_aeolus_l2a",
        "files": files,
    }
    with open(output_path, "wb") as f:
        tomli_w.dump(obsgroup, f)
    click.echo(f"Wrote info about {len(files)} files to {output_path}")


if __name__ == "__main__":
    create_aeolus_l2a_obsgroup()



from pathlib import Path

from wrf_ensembly.external import ExternalProcess


def concatenate(
    ncrcat_cmd: str, input_files: list[Path], output_file: Path, args: list[str] = []
) -> ExternalProcess:
    """
    Concatenate a set of netCDF files using NCO
    You need to execute the returned object to run the command using external.run()

    Args:
        ncrcat_cmd: Path to the ncrcat command to use
        input_files: List of files to concatenate
        output_file: Where to write the output
        args: Additional arguments to pass to ncrcat (e.g., compression)
    """

    output_file.parent.mkdir(parents=True, exist_ok=True)

    return ExternalProcess(
        [
            *ncrcat_cmd.split(" "),
            "-4",
            "-O",
            *args,
            *[str(x.resolve()) for x in input_files],
            str(output_file.resolve()),
        ]
    )



import re
from pathlib import Path
from typing import Union


def convert_to_dtype(
    x: list[str] | str,
) -> list[int | float | str | bool | list] | int | float | str | bool:
    """
    Convert a namelist value to the appropriate data type.
    Can handle integers and floats, and lists of either.
    If the value is a string, it is returned as-is.
    """
    if isinstance(x, list):
        return [convert_to_dtype(v) for v in x]
    x = x.strip()
    if x == ".true.":
        return True
    if x == ".false.":
        return False
    if re.match(r"^[Ff]alse$", x):
        return False
    if re.match(r"^-?[0-9]+$", x):
        return int(x)
    if re.match(r"^-?[0-9]+\.[0-9]*e?[-+]?[0-9]*$", x):
        return float(x)
    if re.match(r"^'.*'$", x):
        return x[1:-1]
    return x


def parse(buf: str) -> dict[str, Union[str, dict, list, str, int, float]]:
    """
    Parse a namelist into a dictionary.

    Args:
        buf: The namelist to parse, in string format.
    """
    groups = {}
    group = None
    for line in buf.splitlines():
        line = line.strip()
        if line.startswith("&"):
            group = line[1:].strip()
            groups[group] = {}
        elif line.startswith("/"):
            group = None
        elif "=" in line:
            key, value = line.split("=", 1)
            key = key.strip()
            value = value.strip()
            if value.endswith(","):
                value = value[:-1]
            if "," in value:
                value = [v.strip() for v in value.split(",")]
            groups[group][key] = convert_to_dtype(value)
    return groups


def read(path: str | Path):
    """
    Read a namelist from a file.

    Args:
        path: The path to read the namelist from.
    """
    with open(path, "r") as f:
        return parse(f.read())


def convert_to_string(
    x: Union[str, bool, int, float, list[Union[str, bool, int, float]]],
) -> str:
    """
    Convert a namelist value to a string, properly formatting it.
    """

    if isinstance(x, list):
        return ", ".join([convert_to_string(v) for v in x])
    if isinstance(x, bool):
        return ".true." if x else ".false."
    if isinstance(x, int) or isinstance(x, float):
        return str(x)
    return f"'{x}'"


def write_namelist(namelist: dict, path: str | Path):
    """
    Write a namelist to a file.

    Args:
        namelist: The namelist to write.
        path: The path to write the namelist to.
    """
    with open(path, "w") as f:
        for group, values in namelist.items():
            print(f"&{group}", file=f)
            for key, value in values.items():
                if value is None:
                    continue
                value = convert_to_string(value)
                print(f"    {key} = {value},", file=f)
            print("/", file=f)



from .database import ExperimentDatabase
from .dataclasses import ExperimentStatus, MemberStatus, RuntimeStatistics
from .experiment import Experiment
from .paths import ExperimentPaths

__all__ = [
    "Experiment",
    "ExperimentDatabase",
    "ExperimentPaths",
    "ExperimentStatus",
    "MemberStatus",
    "RuntimeStatistics",
]



import os
import sys
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Optional

import click
import netCDF4

from wrf_ensembly import experiment, utils
from wrf_ensembly.click_utils import GroupWithStartEndPrint, pass_experiment_path
from wrf_ensembly.console import logger


@click.group(name="ensemble", cls=GroupWithStartEndPrint)
def ensemble_cli():
    pass


@ensemble_cli.command()
@pass_experiment_path
def setup(experiment_path: Path):
    """
    Copies initial/boundary conditions for each member.
    """

    logger.setup("ensemble-setup", experiment_path)
    exp = experiment.Experiment(experiment_path)

    first_cycle = exp.cycles[0]
    logger.info(f"Configuring members for cycle 0: {str(first_cycle)}")

    for i in range(exp.cfg.assimilation.n_members):
        member_dir = exp.paths.member_path(i)

        # First check if there are member-specific IC/BC files, otherwise use the same
        # for all members
        ic_file = (
            exp.paths.data_icbc
            / f"member_{i:02d}"
            / f"wrfinput_d01_member_{i:02d}_cycle_0"
        )
        if ic_file.exists():
            logger.info(f"Member {i}: Using member-specific IC file {ic_file}")
        else:
            ic_file = exp.paths.data_icbc / "wrfinput_d01_cycle_0"

        bc_file = (
            exp.paths.data_icbc
            / f"member_{i:02d}"
            / f"wrfbdy_d01_member_{i:02d}_cycle_0"
        )
        if bc_file.exists():
            logger.info(f"Member {i}: Using member-specific BC file {bc_file}")
        else:
            bc_file = exp.paths.data_icbc / "wrfbdy_d01_cycle_0"

        # Copy initial and boundary conditions
        utils.copy(ic_file, member_dir / "wrfinput_d01")
        logger.info(f"Member {i}: Copied wrfinput_d01")

        utils.copy(bc_file, member_dir / "wrfbdy_d01")
        logger.info(f"Member {i}: Copied wrfbdy_d01_cycle_0")


@ensemble_cli.command()
@click.argument(
    "other-experiment", type=click.Path(dir_okay=True, file_okay=False, path_type=Path)
)
@click.option(
    "--cycle", type=int, required=True, help="Which cycle to use for initialisation"
)
@pass_experiment_path
def setup_from_other_experiment(
    experiment_path: Path, other_experiment: Path, cycle: int
):
    """
    Setup the ensemble using a forecast from another experiment.

    The usecase for this command is having a control experiment that starts earlier, for
    spin-up reasons. You can initialise a second experiment from a mid-point and work
    forwards.

    The other experiment must have the same cycle setup (start/end dates, output interval,
    boundary conditions interval) and the same domain. The forecast files for the requested cycle
    must be available in the scratch directory.
    """

    logger.setup("ensemble-setup-from-other-experiment", experiment_path)
    exp = experiment.Experiment(experiment_path)

    logger.info(f"Opening second experiment at {other_experiment}")
    other_exp = experiment.Experiment(other_experiment.resolve())

    # Check if the domain and time control setups are the same
    res = exp.cfg.domain_control.is_equal(other_exp.cfg.domain_control)
    if type(res) is str:
        logger.error(f"Field is not equal in [domain_control]: {res}")
        sys.exit(1)
    res = exp.cfg.time_control.is_equal(other_exp.cfg.time_control)
    if type(res) is str:
        logger.error(f"Field is not equal in [time_control]: {res}")
        sys.exit(1)
    if exp.cfg.time_control.cycles != {} or other_exp.cfg.time_control.cycles != {}:
        logger.warning(
            "[time_control.cycles] is defined in at least one of the experiments. Be careful!"
        )
    if exp.cfg.assimilation.n_members != other_exp.cfg.assimilation.n_members:
        logger.error("Experiments do not have the same amount of members")
        sys.exit(1)

    # Link other experiments IC/BC directory to current exp.
    icbc_dir = exp.paths.data_icbc
    logger.info(
        f"Removing current IC/BC directory {icbc_dir} and linking to {other_exp.paths.data_icbc}"
    )
    if icbc_dir.is_dir():
        icbc_dir.rmdir()
    elif icbc_dir.is_symlink():
        icbc_dir.unlink()
    icbc_dir.symlink_to(other_exp.paths.data_icbc, target_is_directory=True)

    # Check if the required forecasts exist in the scratch directory & link them in the current experiment
    cycle_end = other_exp.cycles[cycle].end
    required_wrfout_filename = f"wrfout_d01_{cycle_end:%Y-%m-%d_%H:%M:%S}"
    logger.info(f"Required forecast filename: {required_wrfout_filename}")

    for i in range(exp.cfg.assimilation.n_members):
        target_scratch = other_exp.paths.scratch_forecasts_path(cycle=cycle, member=i)
        required_wrfout_file = (target_scratch / required_wrfout_filename).resolve()
        if not required_wrfout_file.exists():
            logger.error(f"Forecast doesn't exist at {required_wrfout_file}")
            sys.exit(1)

        new_dir = exp.paths.scratch_forecasts_path(cycle=cycle, member=i)
        new_dir.mkdir(exist_ok=True, parents=True)

        symlink_loc = new_dir / required_wrfout_filename
        if symlink_loc.exists() and not symlink_loc.is_symlink():
            logger.error(
                f"File already exists at {symlink_loc} and is not a symlink. Too scared to replace."
            )
            sys.exit(1)
        symlink_loc.unlink(missing_ok=True)
        logger.info(f"Linking {symlink_loc} to {required_wrfout_file}")
        symlink_loc.symlink_to(required_wrfout_file)

        # Set member as advanced
        exp.members[i].advanced = True

    # Update experiment status & metadata
    exp.current_cycle_i = cycle
    exp.save_status_to_db()

    logger.info(f"Linked to {other_exp}, cycle = {cycle}.")
    logger.info("Use the `cycle` command to advance this experiment to the next cycle")


@ensemble_cli.command()
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many files to process in parallel",
)
@pass_experiment_path
def generate_perturbations(experiment_path: Path, jobs: Optional[int]):
    """Generates perturbations for all experiment cycles"""

    logger.setup("ensemble-generate-perturbations", experiment_path)
    exp = experiment.Experiment(experiment_path)

    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")
    logger.info(
        f"Applying perturbations every cycle: {exp.cfg.perturbations.apply_perturbations_every_cycle}"
    )

    if exp.cfg.perturbations.apply_perturbations_every_cycle:
        logger.info("Generating perturbations for all cycles...")
        with ProcessPoolExecutor(max_workers=jobs) as executor:
            res = executor.map(exp.generate_perturbations, range(len(exp.cycles)))
            for _ in res:
                pass
    else:
        logger.info("Generating perturbations for first cycle only...")
        exp.generate_perturbations(0)


@ensemble_cli.command()
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many files to process in parallel",
)
@pass_experiment_path
def apply_perturbations(experiment_path: Path, jobs: Optional[int]):
    """
    Applies perturbations to the initial conditions of the current cycle.
    Make sure to update the boundary conditions afterwards!
    """

    logger.setup("ensemble-apply-perturbations", experiment_path)
    exp = experiment.Experiment(experiment_path)

    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")

    with ProcessPoolExecutor(max_workers=jobs) as executor:
        res = executor.map(
            exp.apply_perturbations, range(exp.cfg.assimilation.n_members)
        )
        for _ in res:
            pass


@ensemble_cli.command()
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many files to process in parallel",
)
@pass_experiment_path
def update_bc(experiment_path: Path, jobs: Optional[int]):
    """
    Runs `update_wrf_bc` for all members to update boundary conditions.
    Use this after you have modified the initial conditions (perts or cycling).
    """

    logger.setup("ensemble-update-bc", experiment_path)
    exp = experiment.Experiment(experiment_path)
    exp.set_dart_environment()

    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")

    with ProcessPoolExecutor(max_workers=jobs) as executor:
        results = executor.map(
            exp.update_bc,
            range(exp.cfg.assimilation.n_members),
        )

        for _ in results:
            pass


@ensemble_cli.command()
@click.option("--member", required=True, type=int, help="Which member to advance")
@click.option(
    "--cores",
    type=int,
    help="Number of cores to use for wrf.exe. ",
)
@pass_experiment_path
def advance_member(
    experiment_path: Path,
    member: int,
    cores: int,
):
    """
    Advances the given MEMBER 1 cycle by running the model

    You can control how many cores to use with --cores. If omitted, will check for
    `SLURM_NTASKS` in the environment and use that. If missing, will use 1 core.
    """

    logger.setup(f"ensemble-advance-member_{member}", experiment_path)
    exp = experiment.Experiment(experiment_path)
    exp.set_wrf_environment()
    if member < 0 or member >= exp.cfg.assimilation.n_members:
        logger.error(f"Member {member} does not exist")
        sys.exit(1)

    # Determine number of cores
    if cores is None:
        if "SLURM_NTASKS" in os.environ:
            cores = int(os.environ["SLURM_NTASKS"])
        else:
            cores = 1
            logger.warning("No --cores no SLURM_NTASKS, will use 1 core!")
    logger.info(f"Using {cores} cores for wrf.exe")

    # Run WRF!
    success = exp.advance_member(member, cores=cores)
    if not success:
        sys.exit(1)


@ensemble_cli.command()
@pass_experiment_path
def filter(experiment_path: Path):
    """
    Runs the assimilation filter for the current cycle
    """

    logger.setup("ensemble-filter", experiment_path)
    exp = experiment.Experiment(experiment_path)
    exp.set_dart_environment()
    exp.filter()


@ensemble_cli.command()
@pass_experiment_path
def analysis(experiment_path: Path):
    """
    Combines the DART output files and the forecast to create the analysis.
    """

    logger.setup("ensemble-analysis", experiment_path)
    exp = experiment.Experiment(experiment_path)

    if not exp.filter_run:
        logger.error("Filter has not been run, cannot run analysis!")
        sys.exit(1)

    cycle_i = exp.current_cycle_i
    cycle = exp.current_cycle

    forecast_dir = exp.paths.scratch_forecasts_path(cycle_i)
    analysis_dir = exp.paths.scratch_analysis_path(cycle_i)
    dart_out_dir = exp.paths.scratch_dart_path(cycle_i)

    # Postprocess analysis files
    for member in range(exp.cfg.assimilation.n_members):
        # Copy forecasts to analysis directory
        wrfout_name = "wrfout_d01_" + cycle.end.strftime("%Y-%m-%d_%H:%M:%S")
        forecast_file = forecast_dir / f"member_{member:02d}" / wrfout_name
        analysis_file = analysis_dir / f"member_{member:02d}" / wrfout_name
        utils.copy(forecast_file, analysis_file)

        dart_file = dart_out_dir / f"dart_member_{member:02d}.nc"
        if not dart_file.exists():
            logger.error(f"Member {member}: {dart_file} does not exist")
            sys.exit(1)

        # Copy the state variables from the dart file to the analysis file
        logger.info(f"Member {member}: Copying state variables from {dart_file}")
        with (
            netCDF4.Dataset(dart_file, "r") as nc_dart,  # type: ignore
            netCDF4.Dataset(analysis_file, "r+") as nc_analysis,  # type: ignore
        ):
            for name in exp.cfg.assimilation.state_variables:
                if name not in nc_dart.variables:
                    logger.warning(f"Member {member}: {name} not in dart file")
                    continue
                logger.info(f"Member {member}: Copying {name}")
                nc_analysis[name][:] = nc_dart[name][:]

            # Add experiment name and current cycle information to attributes
            # TODO Standardize this somehow? We must add metadata to all files!
            nc_analysis.experiment_name = exp.cfg.metadata.name
            nc_analysis.current_cycle = cycle_i
            nc_analysis.cycle_start = cycle.start.strftime("%Y-%m-%d_%H:%M:%S")
            nc_analysis.cycle_end = cycle.end.strftime("%Y-%m-%d_%H:%M:%S")

    # Update experiment status
    exp.analysis_run = True
    exp.save_status_to_db()


@ensemble_cli.command()
@click.option(
    "--use-forecast",
    is_flag=True,
    help="Cycle with the latest forecast instead of the analysis",
)
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many files to process in parallel",
)
@pass_experiment_path
def cycle(experiment_path: Path, use_forecast: bool, jobs: Optional[int]):
    """
    Prepares the experiment for the next cycle by copying the cycled variables from the analysis
    to the initial conditions and preparing the namelist.
    """

    logger.setup("cycle", experiment_path)
    exp = experiment.Experiment(experiment_path)

    if not exp.all_members_advanced:
        logger.error("Not all members have advanced to the next cycle, cannot cycle!")
        sys.exit(1)
    if not use_forecast and not exp.analysis_run:
        logger.error(
            "Analysis step is not done for this cycle, either run it or use --use-forecast to cycle w/ the latest forecast"
        )
        sys.exit(1)

    if use_forecast:
        logger.warning("Cycling using the latest forecast")

    cycle_i = exp.current_cycle_i
    next_cycle_i = cycle_i + 1

    if next_cycle_i >= len(exp.cycles):
        logger.error(f"Experiment is finished! No cycle {next_cycle_i}")
        sys.exit(1)

    # Determine job count
    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")

    with ProcessPoolExecutor(max_workers=jobs) as executor:
        results = executor.map(
            exp.cycle_member,
            range(exp.cfg.assimilation.n_members),
            [use_forecast] * exp.cfg.assimilation.n_members,
        )

        for _ in results:
            pass

    # Update experiment status
    exp.set_next_cycle()
    exp.save_status_to_db()



from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import pandas as pd
import cartopy.crs as ccrs


def plot_observation_locations_on_map(
    observations: pd.DataFrame,
    proj: ccrs.Projection | None,
    domain_bounds: tuple[float, float, float, float] | None = None,
    fig_kwargs: dict = {},
    subplot_kwargs: dict = {},
    ax_kwargs: dict = {},
) -> Figure:
    """
    Plot the locations of a set of observations on a map

    It will use the columns `latitude`, `longitude` from the DataFrame. If there is
    also columns for `instrument` and `quantity`, it will use them for a legend and to
    color the points.

    Args:
        observations: DataFrame with columns `latitude`, `longitude`, optionally
            `instrument` and `quantity`.
        proj: Cartopy projection to use for the map. If None, will use PlateCarree.
        domain_bounds: If given, will set the map extent to these bounds (min_x, max_x,
            min_y, max_y). If None, the extent will be set automatically to contain the
            observations. It must be in `proj` if provided, otherwise in PlateCarree.
        fig_kwargs: Additional keyword arguments to pass to `plt.figure()`
        subplot_kwargs: Additional keyword arguments to pass to `plt.subplot()`
        ax_kwargs: Additional keyword arguments to pass to `ax.scatter()`

    Returns:
        A matplotlib Figure object with the plot.
    """

    if "latitude" not in observations.columns:
        raise ValueError("DataFrame must have a 'latitude' column")
    if "longitude" not in observations.columns:
        raise ValueError("DataFrame must have a 'longitude' column")

    fig, ax = plt.subplots(
        subplot_kw={"projection": proj or ccrs.PlateCarree(), **subplot_kwargs},
        figsize=(10, 8),
        **fig_kwargs,
    )
    ax.coastlines()

    if "instrument" in observations.columns and "quantity" in observations.columns:
        for (instrument, quantity), group in observations.groupby(
            ["instrument", "quantity"]
        ):
            label = f"{instrument} - {quantity}"
            ax.scatter(
                group["longitude"],
                group["latitude"],
                label=label,
                transform=ccrs.PlateCarree(),
                s=0.1,
                **ax_kwargs,
            )

        # Create a nice legend outside the plot
        ax.legend(
            loc="lower center",
            bbox_to_anchor=(1, 0.5),
            title="Instrument - Quantity",
            fontsize="small",
            title_fontsize="medium",
            frameon=False,
        )

    else:
        ax.scatter(
            observations["longitude"],
            observations["latitude"],
            transform=ccrs.PlateCarree(),
            s=0.1,
            **ax_kwargs,
        )

    if domain_bounds is not None:
        ax.set_extent(domain_bounds, crs=proj or ccrs.PlateCarree())

    return fig



"""Observation format converters for WRF-Ensembly."""

from .aeronet import aeronet as aeronet_cli
from .remotap_spexone import remotap_spexone as remotap_spexone_cli

__all__ = ["aeronet_cli", "remotap_spexone_cli"]



import re
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path

import click
import netCDF4
import numpy as np
import pandas as pd
import tomli_w


def parse_spex_date(
    utc_date: np.ndarray, fraction_of_day: np.ndarray
) -> pd.DatetimeIndex:
    """
    SPEX dates are provided in a UTC date as an integer and a float representing the fraction of day.
    This function converts them to pandas timestamps.
    """

    utcdate_str = np.array([str(int(date)) for date in utc_date])
    date = pd.to_datetime(utcdate_str, format="%Y%m%d")
    date = date + pd.to_timedelta(fraction_of_day, unit="D")

    return date


def process_spex_file(nc_path: Path) -> tuple[pd.Timestamp, pd.Timestamp] | None:
    """
    Get the start/end time of a SPEX file
    """

    with netCDF4.Dataset(nc_path) as nc:  # type: ignore
        try:
            utcdate = nc["/geolocation_data/utc_date"][:].flatten()
            fracday = nc["/geolocation_data/fracday"][:].flatten()

            # Mask out fill values for utcdate (29990101)
            mask = utcdate != 29990101

            timestamps = parse_spex_date(utcdate[mask], fracday[mask])
        except pd.errors.OutOfBoundsDatetime as ex:
            print(f"File {nc_path} has invalid dates!" + str(ex))
            return None
        start, end = timestamps.min(), timestamps.max()

    return start, end


@click.command()
@click.argument("dir_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(writable=True, path_type=Path))
@click.option(
    "--glob", "-g", default="*.nc", help="Glob pattern filter files/directories"
)
def create_remotap_spexone_obsgroup(dir_path: Path, output_path: Path, glob: str):
    """
    Creates the observation group file for the given directory of AEOLUS L2B DBL files
    """

    # Ensure the output path is writable
    output_path.touch(exist_ok=True)

    # Sample file name:  PACE_SPEXONE.20240226T013118.L2.AER_OCEAN_REMOTAP.nc
    matcher = re.compile(r"PACE_SPEXONE.*AER_(?:LAND|OCEAN)_REMOTAP\.nc")

    potential_files = [
        p for p in dir_path.rglob(glob) if p.is_file() and matcher.match(p.name)
    ]

    with ProcessPoolExecutor() as executor:
        results = list(executor.map(process_spex_file, potential_files))

    # Create the structure for the TOML file
    files = []
    for nc_path, result in zip(potential_files, results):
        if result is None:
            print(f"File {nc_path} has invalid dates!")
            continue
        start, end = result
        start = start.strftime("%Y%m%dT%H%M%S")
        end = end.strftime("%Y%m%dT%H%M%S")
        files.append(dict(path=str(nc_path.resolve()), start_date=start, end_date=end))

    if not files:
        click.echo("No files found!")
        return

    # Sort files by start date, easier to browse the file this way
    files = sorted(files, key=lambda f: f["start_date"])

    obsgroup = {
        "kind": "RemoTAP-SPEXone",
        "converter": "/path/to/remotap_spexone_converter",
        "files": files,
    }
    with open(output_path, "wb") as f:
        tomli_w.dump(obsgroup, f)
    click.echo(f"Wrote info about {len(files)} files to {output_path}")


if __name__ == "__main__":
    create_remotap_spexone_obsgroup()



import itertools
import os
import shutil
import string
import time
import zipfile
from contextlib import contextmanager
from pathlib import Path
from typing import Optional

from wrf_ensembly.console import logger


def rm_tree(p: Path):
    """
    Removes a directory tree recursively.
    Only handles files, symbolic links, and directories.

    Args:
        p: Path to remove
    """

    if p.is_file() or p.is_symlink():
        p.unlink()
    else:
        for child in p.iterdir():
            rm_tree(child)
        p.rmdir()


def int_to_letter_numeral(i: int) -> str:
    """
    Converts the given integer to a letter numeral, This is used by WPS when ungribing files (e.g., GRIBFILE.AAB), ...

    The first file (1) would be A, the second would be B, and so on. After Z, the next file would be BA, then BB, and so on.
    """

    if i < 1 or i > 17576:
        raise ValueError("i must be between 1 and 704 (AAA and ZZZ)")

    letters = list(itertools.product(string.ascii_uppercase, repeat=3))[i - 1]

    return "".join(letters).rjust(3, "A")


def copy(src: Path, dest: Path, ensure_dest_parent_exists=True):
    """
    Copies the file from `src` to `dest` using `shutil.copy` and logs the operation.

    Args:
        src: Source file
        dest: Destination file.
        ensure_dest_parent_exists: Whether to create the parent directory of `dest` if it doesn't exist.
    """

    logger.debug(f"Copying {src} to {dest}")
    if ensure_dest_parent_exists:
        dest.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(src, dest)


def move(src: Path, dest: Path, ensure_dest_parent_exists=True):
    """
    Moves the file from `src` to `dest` using `shutil.move` and logs the operation.

    Args:
        src: Source file
        dest: Destination file.
        ensure_dest_parent_exists: Whether to create the parent directory of `dest` if it doesn't exist.
    """

    logger.debug(f"Moving {src} to {dest}")
    if ensure_dest_parent_exists:
        dest.parent.mkdir(parents=True, exist_ok=True)
    shutil.move(src, dest)


@contextmanager
def atomic_binary_open(path: Path, mode="wb"):
    """
    Opens a file in atomic mode, which means that the file is first opened in a temporary
    location and then moved to the final location after the file is closed.

    This function will attempt to create the temp file in the same directory but with a
    random prefix. So the process/user must have the appropriate permissions to create
    files in the same directory.

    Args:
        path: Path to the file
        mode: Mode to open the file in, default 'wb'
    """

    # Create a temp file
    tmp_file = path.parent / (path.name + ".tmp")
    with open(tmp_file, mode) as f:
        yield f

    # If the target exists, remove it
    if path.exists():
        path.unlink()

    # Move the temp file to the target location
    tmp_file.rename(path)


def filter_none_from_dict(dict: dict) -> dict:
    """
    Returns a new dictionary that does not contain any keys with a value of None.

    Args:
        dict: Dictionary to filter
    """

    return {k: v for k, v in dict.items() if v is not None}


def seconds_to_pretty_hours(seconds: int | float) -> str:
    """
    Converts seconds to hours minutes in the `HHh MMm` format
    """

    hours = seconds // 3600
    minutes = (seconds % 3600) // 60

    return f"{hours:.0f}h {minutes:02.0f}m"


def bool_to_console_str(b: bool):
    """
    Converts a boolean to a string that can be printed to the console.
    """

    return "✅" if b else "❌"


class LockFile:
    """
    A simple lock file that can be used to prevent multiple processes from running the same code at the same time. It is implemented by creating a second file with the .lock suffix.

    Args:
        path: Path to the lock file
    """

    path: Path
    """The file to lock"""

    lockfile: Path
    """Path to the lockfile (.lock)"""

    timeout: int
    """Max amount of seconds to wait for a lockfile to clear"""

    pooling_interval = 5
    """How often to pool an existing lockfile"""

    def __init__(self, path: Path, timeout=10 * 60):
        self.path = path
        self.lockfile = path.with_suffix(".lock")
        self.timeout = timeout

    def __enter__(self):
        waiting_for = 0

        pooling = True
        while pooling:
            if self.lockfile.exists():
                logger.debug(f"Waiting for lockfile to be removed: {self.lockfile}")
                time.sleep(self.pooling_interval)
                waiting_for += self.pooling_interval
            else:
                try:
                    logger.debug(f"Creating lockfile: {self.lockfile}")
                    self.lockfile.touch(exist_ok=False)
                    pooling = False
                except FileExistsError:
                    logger.debug(
                        f"Lockfile was created by another process: {self.lockfile}"
                    )
                    time.sleep(self.pooling_interval)
                    waiting_for += self.pooling_interval

            if waiting_for > self.timeout:
                raise TimeoutError(
                    f"Lockfile was not removed after {self.timeout} seconds: {self.lockfile}"
                )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.debug(f"Releasing lockfile: {self.lockfile}")

        if exc_val is not None:
            logger.error(f"An exception occurred: {exc_val}")
        if self.lockfile.is_file():
            logger.debug(f"Removing lockfile: {self.lockfile}")
            self.lockfile.unlink(missing_ok=True)


def determine_jobs(job_param: Optional[int]) -> int:
    """
    Returns how many jobs you should run in parallel. It is determined by:
    - If `job_param` is not None and a positive number, it is returned.
    - If `job_param` is None, then SLURM_NTASKS is used from the environment if it exists.
    - If `job_param` is None and SLURM_NTASKS is not set, then 1 is returned and a warning is logged.

    Args:
        job_param: Number of jobs to run in parallel

    Returns:
        Number of jobs to run in parallel
    """
    if job_param is not None and job_param > 0:
        return job_param

    slurm_ntasks = os.getenv("SLURM_NTASKS")
    if slurm_ntasks is not None:
        try:
            return int(slurm_ntasks)
        except ValueError:
            logger.warning(f"Invalid SLURM_NTASKS value: {slurm_ntasks}")

    logger.warning(
        "job_param is None and SLURM_NTASKS is not set. Defaulting to 1 job."
    )
    return 1


def zip_files(filelist: list[Path], output: Path):
    """
    Puts all files from `filelist` into a zip archive at `output`.
    """

    with zipfile.ZipFile(output, "w") as zipf:
        for file in filelist:
            zipf.write(file, file.name)



"""Converter for AERONET CSV-ish files to WRF-Ensembly Observation format."""

from pathlib import Path
from typing import List

import click
import numpy as np
import pandas as pd

from wrf_ensembly.observations import io as obs_io


def convert_aeronet(
    path: Path, quantities: list[str] = ["AOD_340", "AOD_500"]
) -> None | pd.DataFrame:
    """Convert an AERONET file to WRF-Ensembly Observation format.

    Args:
        path: Path to the AERONET tabular file.
        quantities: List of quantities to extract from the file. Default is ["AOD_550nm"].

    Returns:
        A pandas DataFrame in WRF-Ensembly Observation format (correct columns etc).
    """

    aeronet_df = pd.read_csv(path, skiprows=6, delimiter=",", encoding="latin-1")
    for col in quantities:
        if col not in aeronet_df.columns:
            raise ValueError(
                f"Requested quantity '{col}' not found in AERONET file columns"
            )

    # Ensure TZ is UTC
    aeronet_df["timestamp"] = pd.to_datetime(
        aeronet_df["Date(dd:mm:yyyy)"] + " " + aeronet_df["Time(hh:mm:ss)"],
        format="%d:%m:%Y %H:%M:%S",
    )
    aeronet_df["timestamp"] = aeronet_df["timestamp"].dt.tz_localize("UTC")
    aeronet_df = aeronet_df.drop(columns=["Date(dd:mm:yyyy)", "Time(hh:mm:ss)"])

    # -999 is used as a missing data marker
    aeronet_df = aeronet_df.replace(-999.0, pd.NA)

    # The AERONET files have many columns (wide format), we need to convert them to long format
    # where each row only has one quantity.
    # AERONET files don't have any uncertainty, so we will set that to NaN for now.
    aeronet_df = aeronet_df.reset_index().melt(
        id_vars=[
            "index",
            "timestamp",
            "Site_Latitude(Degrees)",
            "Site_Longitude(Degrees)",
        ],
        value_vars=quantities,
        var_name="quantity",
        value_name="value",
    )

    # Drop rows without values
    aeronet_df = aeronet_df.dropna(subset=["value"]).reset_index(drop=True)
    if aeronet_df.empty:
        return None  # Nothing to do

    # Ensure the value column is float
    aeronet_df["value"] = aeronet_df["value"].astype(float)

    # Rename existing columns to the correct names, assign the rest
    aeronet_df = aeronet_df.rename(
        columns={
            "Site_Latitude(Degrees)": "latitude",
            "Site_Longitude(Degrees)": "longitude",
            "timestamp": "time",
        }
    )
    aeronet_df["instrument"] = "AERONET"
    aeronet_df["z"] = 0.0
    aeronet_df["z_type"] = "surface"
    aeronet_df["value_uncertainty"] = pd.NA
    aeronet_df["qc_flag"] = 0  # No QC available, so set to "good"
    aeronet_df["orig_coords"] = aeronet_df.apply(
        lambda row: {
            "indices": np.array((row["index"],), dtype=int),
            "shape": np.array((aeronet_df.shape[0],), dtype=int),
            "names": np.array(("row",), dtype=object),
        },
        axis=1,
    )
    aeronet_df["orig_filename"] = path.name
    aeronet_df["metadata"] = pd.NA  # No additional metadata available

    aeronet_df = aeronet_df.drop(columns=["index"])

    # Sort columns as defined in the schema, do the sanity check
    aeronet_df = aeronet_df[obs_io.REQUIRED_COLUMNS]
    obs_io.validate_schema(aeronet_df)

    return aeronet_df


@click.command()
@click.argument("input_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(path_type=Path))
@click.option(
    "--quantities",
    multiple=True,
    default=["AOD_380nm", "AOD_500nm"],
    help="Quantities to extract from the AERONET file. Can be specified multiple times.",
)
def aeronet(input_path: Path, output_path: Path, quantities: List[str]):
    """Convert AERONET CSV file to WRF-Ensembly observation format.

    INPUT_PATH: Path to the AERONET CSV file
    OUTPUT_PATH: Path where to save the converted observations (will be saved as parquet)
    """

    print(f"Converting AERONET file: {input_path}")
    print(f"Output path: {output_path}")
    print(f"Quantities: {', '.join(quantities)}")

    # Convert the data
    converted_df = convert_aeronet(input_path, list(quantities))
    if converted_df is None or converted_df.empty:
        print("No observations found in the input file, aborting")
        return

    # Save to output path as parquet
    obs_io.write_obs(converted_df, output_path)

    print(f"Successfully converted {len(converted_df)} observations")
    print(f"Saved to: {output_path}")



from pathlib import Path
from typing import Optional

from wrf_ensembly.config import Config


class ExperimentPaths:
    """
    Paths to the different directories of an experiment
    """

    def __init__(self, experiment_path: Path, cfg: Config):
        self.experiment_path = experiment_path.resolve()
        self.work_path = experiment_path / "work"
        self.ensemble_path = self.work_path / "ensemble"
        self.jobfiles = experiment_path / "jobfiles"

        # Data directories
        self.data = experiment_path / "data"
        self.data_icbc = self.data / "initial_boundary"
        self.data_forecasts = self.data / "forecasts"
        self.data_analysis = self.data / "analysis"
        self.data_diag = self.data / "diagnostics"

        self.obs = experiment_path / "obs"
        self.obs_temp = self.obs / "temp"  # Temporary files during processing
        self.obs_db = experiment_path / "observations.duckdb"

        self.plots = experiment_path / "plots"

        # Work directories
        self.work = experiment_path / "work"
        self.work_wrf = self.work / "WRF"
        self.work_wps = self.work / "WPS"
        self.work_ensemble = self.work / "ensemble"
        self.member_paths = [
            self.member_path(i) for i in range(cfg.assimilation.n_members)
        ]

        # Preprocessing
        self.work_preprocessing = self.work / "preprocessing"
        self.work_preprocessing_wrf = self.work_preprocessing / "WRF"
        self.work_preprocessing_wps = self.work_preprocessing / "WPS"

        # Logs
        self.logs = experiment_path / "logs"
        self.logs_slurm = self.logs / "slurm"

        # Scratch
        self.scratch = cfg.directories.scratch_root
        if not self.scratch.is_absolute():
            self.scratch = experiment_path / self.scratch
        self.scratch_forecasts = self.scratch / "forecasts"
        self.scratch_analysis = self.scratch / "analysis"
        self.scratch_dart = self.scratch / "dart"

    def create_directories(self):
        """Creates all required directories"""
        self.obs.mkdir()
        self.work.mkdir()
        self.work_preprocessing.mkdir()
        self.jobfiles.mkdir()
        self.plots.mkdir()

        self.data.mkdir()
        self.data_analysis.mkdir()
        self.data_forecasts.mkdir()
        self.data_icbc.mkdir()
        self.data_diag.mkdir()

        self.scratch.mkdir()
        self.scratch_forecasts.mkdir()
        self.scratch_analysis.mkdir()
        self.scratch_dart.mkdir()

        self.logs.mkdir(exist_ok=True)
        self.logs_slurm.mkdir()

    def member_path(self, i: int) -> Path:
        """
        Get the work directory for given ensemble member
        """
        return self.ensemble_path / f"member_{i:02d}"

    def forecast_path(
        self, cycle: Optional[int] = None, member: Optional[int] = None
    ) -> Path:
        if cycle is None:
            return self.data_forecasts
        if member is None:
            return self.data_forecasts / f"cycle_{cycle:03d}"
        return self.data_forecasts / f"cycle_{cycle:03d}" / f"member_{member:02d}"

    def analysis_path(self, cycle: Optional[int] = None) -> Path:
        if cycle is None:
            return self.data_analysis
        return self.data_analysis / f"cycle_{cycle:03d}"

    def scratch_forecasts_path(
        self, cycle: Optional[int] = None, member: Optional[int] = None
    ) -> Path:
        if cycle is None:
            return self.scratch_forecasts
        if member is None:
            return self.scratch_forecasts / f"cycle_{cycle:03d}"
        return self.scratch_forecasts / f"cycle_{cycle:03d}" / f"member_{member:02d}"

    def scratch_analysis_path(self, cycle: Optional[int] = None) -> Path:
        if cycle is None:
            return self.scratch_analysis
        return self.scratch_analysis / f"cycle_{cycle:03d}"

    def scratch_dart_path(self, cycle: Optional[int] = None) -> Path:
        if cycle is None:
            return self.scratch_dart
        return self.scratch_dart / f"cycle_{cycle:03d}"



import concurrent.futures
import re
import sys
from itertools import chain
from pathlib import Path
from typing import Optional

import click
import xarray as xr

from wrf_ensembly import experiment, external, nco, processors
from wrf_ensembly import statistics as stats
from wrf_ensembly import utils
from wrf_ensembly.click_utils import GroupWithStartEndPrint, pass_experiment_path
from wrf_ensembly.console import logger
from wrf_ensembly.processors import ProcessingContext, process_file_with_pipeline


@click.group(name="postprocess", cls=GroupWithStartEndPrint)
def postprocess_cli():
    pass


@postprocess_cli.command()
@pass_experiment_path
def print_variables_to_keep(experiment_path: Path):
    """
    Prints which variables will be kept in the wrfout files after postprocessing.
    This is useful to check if the variables you want to keep are actually kept or check
    how the regex filters are applied.

    The variables are defined in the `PostprocessConfig:variables_to_keep` variable of the
    configuration file.

    Cycle 0 output for member 0 must exist for this command to work.
    """

    logger.setup("postprocess-print_variables_to_keep", experiment_path)
    exp = experiment.Experiment(experiment_path)

    if not exp.cfg.postprocess.variables_to_keep:
        logger.warning("`variables_to_keep` is not set in the config, exiting")
        sys.exit(0)

    # Find a sample forecast file, cycle 0, member 0
    scratch_forecast_dir = exp.paths.scratch_forecasts_path(0)
    forecast_files = list(scratch_forecast_dir.rglob("member_00/wrfout*"))
    if len(forecast_files) == 0:
        logger.error("No forecast files found, exiting")
        sys.exit(1)

    # Use the first file
    forecast_file = forecast_files[0]
    logger.info(f"Using {forecast_file} as sample file")
    ds = xr.open_dataset(forecast_file, decode_times=False)

    # Add variables that would normally be generated by xwrf_post
    for new_var in [
        "air_potential_temperature",
        "air_pressure",
        "wind_east",
        "wind_north",
        "air_density",
    ]:
        ds[new_var] = xr.zeros_like(ds["THM"])
    ds = ds.rename({"XTIME": "t"})

    filters = [re.compile(x) for x in exp.cfg.postprocess.variables_to_keep]
    logger.info("Variables in output file:")
    for var in ds.data_vars:
        for f in filters:
            if f.match(str(var)):
                logger.info(f"{f} -> {var}: ({ds[var].dims}) ({ds[var].dtype})")
                break


@postprocess_cli.command()
@click.option(
    "--cycle",
    type=int,
    help="Cycle to compute statistics for. Will compute for current cycle if missing.",
)
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many files to process in parallel",
)
@pass_experiment_path
def process_pipeline(experiment_path: Path, cycle: Optional[int], jobs: Optional[int]):
    """
    Apply the configured data processor pipeline to output files.

    This command replaces apply-scripts with a more efficient plugin-based approach
    that processes data in memory using a configurable pipeline of DataProcessor instances.
    """

    logger.setup("postprocess-process-pipeline", experiment_path)
    exp = experiment.Experiment(experiment_path)

    if cycle is None:
        cycle = exp.current_cycle_i

    logger.info(f"Cycle: {exp.cycles[cycle]}")

    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")

    try:
        pipeline = processors.create_pipeline_from_config(
            exp.cfg.postprocess.processors
        )
        logger.info(
            f"Created pipeline with processors: {[p.name for p in pipeline.processors]}"
        )
    except Exception as e:
        logger.error(f"Failed to create processor pipeline: {e}")
        sys.exit(1)

    scratch_forecast_dir = exp.paths.scratch_forecasts_path(cycle)
    scratch_analysis_dir = exp.paths.scratch_analysis_path(cycle)

    # Clean any old _post files first
    for f in scratch_forecast_dir.rglob("wrfout*_post"):
        logger.debug(f"Removing old file {f}")
        f.unlink()
    for f in scratch_analysis_dir.rglob("wrfout*_post"):
        logger.debug(f"Removing old file {f}")
        f.unlink()

    files_to_process = list(scratch_analysis_dir.rglob("member_*/wrfout*")) + list(
        scratch_forecast_dir.rglob("member_*/wrfout*")
    )
    # Filter out any _post files that might still exist, just in case?
    files_to_process = [f for f in files_to_process if not f.name.endswith("_post")]

    logger.info(f"Found {len(files_to_process)} files to process")

    if len(files_to_process) == 0:
        logger.warning("No files found to process")
        return

    # Create processor arguments for each file
    process_args = []
    for file_path in files_to_process:
        # Extract metadata from file path
        member_match = re.search(r"member_(\d+)", str(file_path))
        member_id = int(member_match.group(1)) if member_match else 0

        # Create output file path with _post suffix
        output_file = file_path.parent / f"{file_path.name}_post"

        context = ProcessingContext(
            member=member_id,
            cycle=cycle,
            input_file=file_path,
            output_file=output_file,
            config=exp.cfg,
        )

        process_args.append(
            (file_path, output_file, exp.cfg.postprocess.processors, context)
        )

    logger.info(f"Processing {len(files_to_process)} files through pipeline")
    with concurrent.futures.ProcessPoolExecutor(max_workers=jobs) as executor:
        futures = [
            executor.submit(process_file_with_pipeline, args) for args in process_args
        ]

        # Wait for all to complete and handle any errors
        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()
            except Exception as e:
                logger.error(f"Pipeline processing failed: {e}")
                sys.exit(1)

    logger.info("Pipeline processing completed successfully")


@postprocess_cli.command()
@click.option(
    "--cycle",
    type=int,
    help="Cycle to compute statistics for. Will compute for all current cycle if missing.",
)
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many files to process in parallel",
)
@pass_experiment_path
def statistics(
    experiment_path: Path,
    cycle: Optional[int],
    jobs: int,
):
    """
    Calculates the ensemble mean and standard deviation from the forecast/analysis files of given cycle.
    This function reads the `*_post` files created by the `wrf_post` and, optionally, `apply_scripts` commands.
    """

    logger.setup("postprocess-statistics", experiment_path)
    exp = experiment.Experiment(experiment_path)

    if cycle is None:
        cycle = exp.current_cycle_i
    logger.info(f"Cycle: {exp.cycles[cycle]}")

    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")

    # Escape hatch when only one member is present
    if exp.cfg.assimilation.n_members == 1:
        logger.warning(
            "Only one member, copying file over unchanged, no standard deviation."
        )

        analysis_path = exp.paths.scratch_analysis_path(cycle) / "member_00"
        forecast_path = exp.paths.scratch_forecasts_path(cycle) / "member_00"
        print(forecast_path)
        for f in chain(
            analysis_path.rglob("wrfout*_post"),
            forecast_path.rglob("wrfout*_post"),
        ):
            print(f)
            target = f.parent.parent / f.name.replace("_post", "_mean")
            utils.copy(f, target)

        return

    # We have to use `stats.compute_ensemble_statistics` once for each output file in the cycle (both forecast and analysis)
    # First, find out how many output files we have by checking the first member
    args: list[tuple[list[Path], tuple[Path, Path]]] = []

    scratch_forecast_dir = exp.paths.scratch_forecasts_path(cycle)
    forecast_filenames = [
        x.name for x in scratch_forecast_dir.rglob("member_00/wrfout*_post")
    ]
    for name in forecast_filenames:
        output_mean_file = scratch_forecast_dir / name.replace("_post", "_mean")
        output_sd_file = scratch_forecast_dir / name.replace("_post", "_sd")

        input_files = [
            scratch_forecast_dir / f"member_{i:02d}/{name}"
            for i in range(exp.cfg.assimilation.n_members)
        ]

        args.append((input_files, (output_mean_file, output_sd_file)))

    # Now do the same for the analysis files
    scratch_analysis_dir = exp.paths.scratch_analysis_path(cycle)
    analysis_filenames = [
        x.name for x in scratch_analysis_dir.rglob("member_00/wrfout*_post")
    ]
    for name in analysis_filenames:
        output_mean_file = scratch_analysis_dir / name.replace("_post", "_mean")
        output_sd_file = scratch_analysis_dir / name.replace("_post", "_sd")

        input_files = [
            scratch_analysis_dir / f"member_{i:02d}/{name}"
            for i in range(exp.cfg.assimilation.n_members)
        ]

        args.append((input_files, (output_mean_file, output_sd_file)))

    logger.info(f"Found {len(args)} filesets to process")

    # Call `compute_ensemble_statistics` for each fileset, in parallel
    with concurrent.futures.ProcessPoolExecutor(max_workers=jobs) as executor:
        futures = []
        for input_files, output_files in args:
            futures.append(
                executor.submit(
                    stats.compute_ensemble_statistics,
                    input_files,
                    output_files[0],
                    output_files[1],
                )
            )

        for res in concurrent.futures.as_completed(futures):
            res.result()

    logger.info("All files processed successfully")


@postprocess_cli.command()
@click.option(
    "--cycle",
    type=int,
    help="Cycle to compute statistics for. Will compute for all current cycle if missing.",
)
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    default=4,
    help="How many NCO commands to execute in parallel",
)
@pass_experiment_path
def concatenate(
    experiment_path: Path,
    cycle: Optional[int],
    jobs: int,
):
    """
    Concatenates all output files (mean and standard deviation) into two files, one for analysis
    and one for forecast. It uses the `_mean` and `_sd` files created by the `statistics` command.
    """

    logger.setup("postprocess-statistics", experiment_path)
    exp = experiment.Experiment(experiment_path)

    if cycle is None:
        cycle = exp.current_cycle_i

    commands = []

    # Prepare compression related arguments
    cmp_args = []
    if len(exp.cfg.postprocess.ppc_filter) > 0:
        cmp_args = ["--ppc", exp.cfg.postprocess.ppc_filter]
    if len(exp.cfg.postprocess.compression_filters) > 0:
        cmp_args.append(f"--cmp={exp.cfg.postprocess.compression_filters}")

    if len(cmp_args) == 0:
        logger.warning("No compression filters set, output files will be uncompressed")
    else:
        logger.info(f"Using compression filters: {' '.join(cmp_args)}")

    # Find all forecast files
    forecast_dir = exp.paths.forecast_path(cycle)
    scratch_forecast_dir = exp.paths.scratch_forecasts_path(cycle)
    forecast_files = sorted(scratch_forecast_dir.rglob("*_mean"))
    if len(forecast_files) > 0:
        commands.append(
            nco.concatenate(
                exp.cfg.postprocess.ncrcat_cmd,
                forecast_files,
                forecast_dir / f"forecast_mean_cycle_{cycle:03d}.nc",
                cmp_args,
            )
        )
    forecast_files = sorted(scratch_forecast_dir.rglob("*_sd"))
    if len(forecast_files) > 0:
        commands.append(
            nco.concatenate(
                exp.cfg.postprocess.ncrcat_cmd,
                forecast_files,
                forecast_dir / f"forecast_sd_cycle_{cycle:03d}.nc",
                cmp_args,
            )
        )

    # Find all analysis files
    analysis_dir = exp.paths.analysis_path(cycle)
    scratch_analysis_dir = exp.paths.scratch_analysis_path(cycle)
    analysis_files = sorted(scratch_analysis_dir.rglob("*_mean"))
    if len(analysis_files) > 0:
        commands.append(
            nco.concatenate(
                exp.cfg.postprocess.ncrcat_cmd,
                analysis_files,
                analysis_dir / f"analysis_mean_cycle_{cycle:03d}.nc",
                cmp_args,
            )
        )
    analysis_files = sorted(scratch_analysis_dir.rglob("*_sd"))
    if len(analysis_files) > 0:
        commands.append(
            nco.concatenate(
                exp.cfg.postprocess.ncrcat_cmd,
                analysis_files,
                analysis_dir / f"analysis_sd_cycle_{cycle:03d}.nc",
                cmp_args,
            )
        )

    # Concatenate per-member if enabled
    if exp.cfg.postprocess.keep_per_member:
        for i in range(exp.cfg.assimilation.n_members):
            forecast_files = list(
                scratch_forecast_dir.glob(f"member_{i:02d}/wrfout*_post")
            )
            commands.append(
                nco.concatenate(
                    exp.cfg.postprocess.ncrcat_cmd,
                    sorted(forecast_files),
                    forecast_dir / f"forecast_member_{i:02d}_cycle_{cycle:03d}.nc",
                    cmp_args,
                )
            )

    failure = False
    logger.info(
        f"Executing {len(commands)} nco commands in parallel, using {jobs} jobs"
    )
    for res in external.run_in_parallel(commands, jobs):
        if res.returncode != 0:
            logger.error(f"nco command failed with exit code {res.returncode}")
            logger.error(res.output)
            failure = True

    if failure:
        logger.error("One or more nco commands failed, exiting")
        sys.exit(1)


@postprocess_cli.command()
@click.option(
    "--cycle",
    type=int,
    help="Cycle to clean up. Will clean for current cycle if missing.",
)
@click.option(
    "--remove-wrfout",
    default=True,
    is_flag=True,
    help="Remove the raw wrfout files",
)
@pass_experiment_path
def clean(experiment_path: Path, cycle: Optional[int], remove_wrfout: bool):
    """
    Clean up the scratch directory for the given cycle. Use after running the other
    postprocessing commands to save disk space.
    """

    logger.setup("postprocess-statistics", experiment_path)
    exp = experiment.Experiment(experiment_path)

    if cycle is None:
        cycle = exp.current_cycle_i

    logger.info(f"Cleaning scratch for cycle {cycle}")

    scratch_dirs = [
        exp.paths.scratch_forecasts_path(cycle),
        exp.paths.scratch_analysis_path(cycle),
    ]
    for dir in scratch_dirs:
        if remove_wrfout:
            for f in dir.rglob("wrfout*"):
                logger.info(f"Removing {f}")
                f.unlink()



import datetime as dt
import shutil
import sys
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Optional

import click
import pandas as pd
from rich.console import Console
from rich.table import Column, Table

from wrf_ensembly import experiment, external, obs_sequence, utils
from wrf_ensembly.click_utils import GroupWithStartEndPrint, pass_experiment_path
from wrf_ensembly.console import logger


@click.group(name="observation-sequence", cls=GroupWithStartEndPrint)
def observation_sequence_cli():
    pass


@observation_sequence_cli.command()
@click.argument("cycle", required=False, type=int)
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many files to process in parallel",
)
@pass_experiment_path
def convert_obs(experiment_path: Path, cycle: Optional[int], jobs: Optional[int]):
    """Converts observation files to DART obs_seq format (does not join per cycle)"""

    logger.setup("observations-convert-obs", experiment_path)
    exp = experiment.Experiment(experiment_path)
    exp.set_dart_environment()

    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")

    # If a cycle is not given, we will convert for all cycles
    cycles = exp.cycles
    if cycle is not None:
        cycles = [c for c in cycles if c.index == cycle]

    # Prepare observation groups for all toml file
    obs_path = exp.paths.obs
    obs_groups = obs_sequence.read_observations(obs_path)
    names = list(obs_groups.keys())

    if len(names) == 0:
        logger.error("No observation groups found!")
        sys.exit(1)

    logger.info(f"Found observation groups: {', '.join(names)}")

    # Gather all files for conversion
    cycle_files = {}
    for c in cycles:
        cycle_files[c.index] = {}

        assimilation_window_start = c.end - dt.timedelta(
            minutes=exp.cfg.assimilation.half_window_length_minutes
        )
        assimilation_window_end = c.end + dt.timedelta(
            minutes=exp.cfg.assimilation.half_window_length_minutes
        )

        for key, obs_group in obs_groups.items():
            cycle_files[c.index][key] = []

            for i, file in enumerate(
                obs_group.get_files_in_window(
                    assimilation_window_start, assimilation_window_end
                )
            ):
                out = obs_path / f"cycle_{c.index}.{key}.{i}.obs_seq"
                cycle_files[c.index][key].append((file.path, out))

    # Print information about the files to convert
    table = Table(
        "Cycle",
        "Window start",
        "Window end",
        *(Column(header=k, justify="right") for k in names),
        title="Observation files to convert per cycle",
    )
    for c in cycles:
        table.add_row(
            str(c.index),
            c.start.strftime("%Y-%m-%d %H:%M"),
            c.end.strftime("%Y-%m-%d %H:%M"),
            *(str(len(cycle_files[c.index][k])) for k in names),
        )
    Console().print(table)

    # Write a dataframe with all files that match each cycle
    rows = []
    for c in cycles:
        for group_name in names:
            for file, out in cycle_files[c.index][group_name]:
                rows.append(
                    {
                        "cycle": c.index,
                        "obs_group": group_name,
                        "file": file.resolve(),
                    }
                )
    df = pd.DataFrame(rows)
    df = df.sort_values(by=["cycle", "obs_group"])
    df = df.reset_index(drop=True)
    df.to_csv(obs_path / "observations_files_used.csv", index=False, header=True)

    # Convert observation files for each cycle in parallel
    commands = []
    for c in cycles:
        for key, obs_group in obs_groups.items():
            for file, out in cycle_files[c.index][key]:
                commands.append(
                    external.ExternalProcess(
                        [*obs_group.converter.split(" "), file, out], cwd=obs_group.cwd
                    )
                )
    for res in external.run_in_parallel(commands, jobs):
        if res.returncode != 0:
            logger.error(f"Failed to convert file: {res.command}")
            logger.error(res.output)
            sys.exit(1)

    logger.info("Conversion complete. Run 'combine-obs' to join files per cycle.")


@observation_sequence_cli.command()
@click.argument("cycle", required=False, type=int)
@click.option(
    "--jobs",
    type=click.IntRange(min=0, max=None),
    help="How many cycles to process in parallel",
)
@pass_experiment_path
def combine_obs(experiment_path: Path, cycle: Optional[int], jobs: Optional[int]):
    """Joins converted observation files into a single obs_seq per cycle"""

    logger.setup("observations-combine-obs", experiment_path)
    exp = experiment.Experiment(experiment_path)
    exp.set_dart_environment()

    jobs = utils.determine_jobs(jobs)
    logger.info(f"Using {jobs} jobs")

    # If a cycle is not given, we will combine for all cycles
    cycles = exp.cycles
    if cycle is not None:
        cycles = [c for c in cycles if c.index == cycle]

    obs_path = exp.paths.obs
    obs_groups = obs_sequence.read_observations(obs_path)
    names = list(obs_groups.keys())

    if len(names) == 0:
        logger.error("No observation groups found!")
        sys.exit(1)

    logger.info(f"Found observation groups: {', '.join(names)}")

    # Gather all files for combining
    cycle_files = {}
    for c in cycles:
        cycle_files[c.index] = {}
        for key in obs_groups.keys():
            # Find all files matching the pattern for this cycle and group
            files = sorted(obs_path.glob(f"cycle_{c.index}.{key}.*.obs_seq"))
            cycle_files[c.index][key] = [(None, f) for f in files]

    # Combine files per cycle
    with ThreadPoolExecutor(max_workers=jobs) as executor:
        futures = []
        for c in cycles:
            cycle_files_list = [
                out for key in names for _, out in cycle_files[c.index][key]
            ]
            if not cycle_files_list:
                continue

            # Filter only for files that exist and are non-empty
            cycle_files_list = [
                f for f in cycle_files_list if f.exists() and f.stat().st_size > 0
            ]

            futures.append(
                executor.submit(
                    obs_sequence.join_obs_seq,
                    exp.cfg,
                    cycle_files_list,
                    obs_path / f"cycle_{c.index}.obs_seq",
                    [v.kind for v in obs_groups.values()],
                )
            )

        for future in futures:
            future.result()

    # Remove temporary files
    for c in cycles:
        for key in names:
            for _, out in cycle_files[c.index][key]:
                out.unlink()

    logger.info("Combine complete.")


@observation_sequence_cli.command()
@click.argument("window_center", required=False, type=click.DateTime())
@click.argument(
    "output_path", required=True, type=click.Path(path_type=Path, writable=True)
)
@click.option("window_length", "-l", "--length", type=int, default=60)
@pass_experiment_path
def prepare_custom_window(
    experiment_path: Path,
    window_center: dt.datetime,
    output_path: Path,
    window_length: int,
):
    """Converts observation files to DART obs_seq format for a custom assimilation window"""

    logger.setup("observations-prepare-custom", experiment_path)
    exp = experiment.Experiment(experiment_path)
    exp.set_dart_environment()

    # Prepare observation groups for all toml file
    obs_path = exp.paths.obs
    obs_groups = obs_sequence.read_observations(obs_path)
    names = list(obs_groups.keys())

    if len(names) == 0:
        logger.error("No observation groups found!")
        sys.exit(1)

    logger.info(f"Found observation groups: {', '.join(names)}")

    # Convert observation files for each cycle
    logger.info("Converting observations for custom assimilation window")
    window_center = window_center.replace(tzinfo=dt.timezone.utc)
    assimilation_window_start = window_center - dt.timedelta(minutes=window_length / 2)
    assimilation_window_end = window_center + dt.timedelta(minutes=window_length / 2)
    logger.info(f"Assimilation window start: {assimilation_window_start.isoformat()}")
    logger.info(f"Assimilation window end: {assimilation_window_end.isoformat()}")

    window_files = []
    for key, obs_group in obs_groups.items():
        logger.info(f"Converting group {key}({obs_group.kind})")

        for i, file in enumerate(
            obs_group.get_files_in_window(
                assimilation_window_start, assimilation_window_end
            )
        ):
            logger.info(f"Converting file {file.path} to obs_seq format")
            out = output_path.parent / f"{key}.{i}.obs_seq"
            obs_group.convert_file(file, out)
            window_files.append(out)

    if len(window_files) == 0:
        logger.warning("No observation files found for this window!")

    # Join files for this group
    logger.info("Joining files...")
    kinds = [v.kind for v in obs_groups.values()]
    obs_sequence.join_obs_seq(exp.cfg, window_files, output_path, kinds)

    # Remove temporary files
    for f in window_files:
        f.unlink()


@observation_sequence_cli.command()
@click.argument(
    "obs_seq_path", required=True, type=click.Path(exists=True, path_type=Path)
)
@click.argument("nc_path", required=True, type=click.Path(path_type=Path))
@pass_experiment_path
def obs_seq_to_nc(experiment_path: Path, obs_seq_path: Path, nc_path: Path):
    """Convert the given obs_seq file to netCDF format"""

    exp = experiment.Experiment(experiment_path)
    exp.set_dart_environment()
    logger.setup("observations-convert-obs-seq", experiment_path)

    obs_sequence.obs_seq_to_nc(exp.cfg.directories.dart_root, obs_seq_path, nc_path)


@observation_sequence_cli.command()
@click.option(
    "--backup/--no-backup",
    is_flag=True,
    default=True,
    help="Whether to backup the original `obs` directory to `obs.tar.gz`. Existing backup is removed.",
)
@pass_experiment_path
def preprocess_for_wrf(experiment_path: Path, backup: bool):
    """
    Runs all obs_seq files through the WRF preprocessing utility to remove observations
    outside the domain and optionally increase obs. errors near the boundary.
    If and how much is this error increase can be configured inside the `observations` config group.
    """

    logger.setup("observations-preprocess-for-wrf", experiment_path)
    exp = experiment.Experiment(experiment_path)
    exp.set_dart_environment()

    # Find a wrfinput file to get the domain
    if exp.cfg.data.per_member_meteorology:
        wrfinput = exp.paths.data_icbc / "member_00" / "wrfinput_d01_member_00_cycle_0"
    else:
        wrfinput = exp.paths.data_icbc / "wrfinput_d01_cycle_0"
    if not wrfinput.exists():
        logger.error(f"wrfinput file not found at {wrfinput}")
        sys.exit(1)

    # Backup the original obs directory by compressing the directory
    if backup:
        target = exp.paths.data / "obs"
        target.unlink(missing_ok=True)

        logger.info(f"Backing up original `obs` directory to {target}")
        shutil.make_archive(str(target), "gztar", str(exp.paths.obs))

    obs_path = exp.paths.obs
    for cycle in exp.cycles:
        obs_seq = obs_path / f"cycle_{cycle.index}.obs_seq"
        if not obs_seq.exists():
            logger.warning(
                f"Skipping cycle {cycle.index} ({obs_seq}) as it does not exist"
            )
            continue

        assimilation_dt = cycle.end
        base_time = dt.datetime(1601, 1, 1, tzinfo=dt.timezone.utc)

        dart_days = (assimilation_dt - base_time).days
        dart_seconds = int(
            (
                assimilation_dt - assimilation_dt.replace(hour=0, minute=0, second=0)
            ).total_seconds()
        )
        logger.debug(f"DART time: {dart_days}, {dart_seconds}")

        # Date must be format in gregorian day, second format
        assimilation_dt = assimilation_dt.replace(tzinfo=dt.timezone.utc)

        logger.info(f"Preprocessing {obs_seq}")
        obs_sequence.preprocess_for_wrf(
            exp.cfg.directories.dart_root,
            wrfinput,
            obs_seq,
            (dart_days, dart_seconds),
            exp.cfg.observations.boundary_width,
            exp.cfg.observations.boundary_error_factor,
            exp.cfg.observations.boundary_error_width,
        )


@observation_sequence_cli.command()
@click.argument("cycle", required=True, type=int)
@click.option(
    "--write-to-file",
    type=click.Path(path_type=Path),
    help="Write filenames to the specified text file (one per line)",
)
@pass_experiment_path
def list_files(experiment_path: Path, cycle: int, write_to_file: Optional[Path]):
    """Lists observation files that match the given cycle's assimilation window"""

    logger.setup("observations-list-files", experiment_path)
    exp = experiment.Experiment(experiment_path)

    # Find the specified cycle
    cycles = [c for c in exp.cycles if c.index == cycle]
    if not cycles:
        logger.error(f"Cycle {cycle} not found!")
        sys.exit(1)

    target_cycle = cycles[0]

    # Prepare observation groups
    obs_path = exp.paths.obs
    obs_groups = obs_sequence.read_observations(obs_path)
    names = list(obs_groups.keys())

    if len(names) == 0:
        logger.error("No observation groups found!")
        sys.exit(1)

    # Calculate assimilation window (same logic as convert_obs)
    assimilation_window_start = target_cycle.end - dt.timedelta(minutes=30)
    assimilation_window_end = target_cycle.end + dt.timedelta(minutes=30)

    logger.info(f"Cycle {cycle} assimilation window:")
    logger.info(f"  Start: {assimilation_window_start.strftime('%Y-%m-%d %H:%M:%S')}")
    logger.info(f"  End: {assimilation_window_end.strftime('%Y-%m-%d %H:%M:%S')}")
    logger.info("")

    # Find matching files for each observation group
    total_files = 0
    all_files = []
    for group_name, obs_group in obs_groups.items():
        matching_files = list(
            obs_group.get_files_in_window(
                assimilation_window_start, assimilation_window_end
            )
        )

        logger.info(
            f"Group '{group_name}' ({obs_group.kind}): {len(matching_files)} files"
        )
        for file in matching_files:
            logger.info(f"  {file.path}")
            all_files.append(str(file.path))

        total_files += len(matching_files)
        if matching_files:
            logger.info("")

    logger.info(f"Total files: {total_files}")

    # Write to file if requested
    if write_to_file:
        write_to_file.parent.mkdir(parents=True, exist_ok=True)
        with open(write_to_file, "w") as f:
            for filepath in all_files:
                f.write(f"{filepath}\n")
        logger.info(f"Filenames written to {write_to_file}")



"""
Functions related to running external commands
"""

import os
import subprocess
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Sequence

from wrf_ensembly.console import logger


@dataclass
class ExternalProcess:
    """Represents a command to run in parallel"""

    command: Sequence[str | Path]
    cwd: Path = Path(os.getcwd())
    log_filename: Optional[str] = None
    stdin: Optional[str] = None


@dataclass
class ExternalProcessResult:
    """Represents a command that has been run in parallel"""

    command: Sequence[str]
    cwd: Path
    returncode: int
    output: str
    log_filename: Optional[str] = None


class ExternalProcessFailed(Exception):
    """Raise this exception when an external command fails and we cannot handle the issue"""

    res: ExternalProcessResult

    def __init__(self, res: ExternalProcessResult):
        self.res = res

    def __str__(self):
        return (
            f"Command {self.res.command} failed with return code {self.res.returncode}"
        )


def run(proc: ExternalProcess):
    """
    Runs a command and returns the output
    """

    command = []
    for c in proc.command:
        if isinstance(c, Path):
            command.append(str(c.resolve()))
        else:
            command.append(c)

    logger.debug(f"Running command: {' '.join(command)}")

    p = subprocess.run(
        command,
        cwd=proc.cwd,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        input=proc.stdin,
    )
    if p.returncode != 0:
        logger.error(f"Command {proc.command} failed with return code {p.returncode}")
        logger.error(p.stdout)

    # Write stdout/err to this command's log directory
    if proc.log_filename is not None:
        logger.write_log_file(proc.log_filename, p.stdout)

    return ExternalProcessResult(
        command=command,
        cwd=proc.cwd,
        returncode=p.returncode,
        output=p.stdout,
    )


def runc(
    command: Sequence[str | Path],
    cwd: Path = Path(os.getcwd()),
    log_filename: Optional[str] = None,
):
    """Brief form of run() in case you don't need the ExternalProcess object"""
    return run(ExternalProcess(command, cwd, log_filename))


def run_in_parallel(
    commands: list[ExternalProcess],
    max_processes: int = 1,
    stop_on_failure: bool = False,
):
    """
    Runs a list of commands in parallel, with a maximum of `max_processes` processes running at the same time.
    If stop_on_failure is True, cancels remaining jobs when the first failure occurs.
    """

    with ThreadPoolExecutor(max_workers=max_processes) as executor:
        futures = [executor.submit(run, command) for command in commands]

        try:
            for future in as_completed(futures):
                res = future.result()
                str_cmd = " ".join(res.command)

                if res.returncode != 0:
                    logger.error(
                        f"Command {str_cmd} failed with return code {res.returncode}"
                    )
                    if stop_on_failure:
                        # Cancel all remaining futures
                        for f in futures:
                            f.cancel()
                        yield res
                        return
                else:
                    logger.debug(
                        f"Command {str_cmd} finished with return code {res.returncode}"
                    )
                yield res
        except KeyboardInterrupt:
            logger.info("KeyboardInterrupt received, cancelling remaining jobs...")
            # Cancel all futures
            for f in futures:
                f.cancel()
            raise


def assert_all_successful(results: list[ExternalProcessResult]):
    """
    Assert that all commands in the list of results were successful
    """

    for res in results:
        if res.returncode != 0:
            sys.exit(1)



import datetime as dt
import os
from pathlib import Path

import netCDF4
import numpy as np
import xarray as xr

from wrf_ensembly import (
    config,
    cycling,
    external,
    obs_sequence,
    perturbations,
    update_bc,
    utils,
    wrf,
)
from wrf_ensembly.console import logger

from .database import ExperimentDatabase
from .dataclasses import MemberStatus, RuntimeStatistics
from .paths import ExperimentPaths
from .observations import ExperimentObservations


class Experiment:
    """
    An ensemble assimilation experiment
    """

    cfg: config.Config
    cycles: list[cycling.CycleInformation]
    current_cycle_i: int
    filter_run: bool
    analysis_run: bool
    paths: ExperimentPaths
    members: list[MemberStatus] = []

    db: ExperimentDatabase
    obs: ExperimentObservations

    def __init__(self, experiment_path: Path):
        self.cfg = config.read_config(experiment_path / "config.toml")
        self.cycles = cycling.get_cycle_information(self.cfg)

        # Initialize database
        db_path = experiment_path / "status.db"
        self.db = ExperimentDatabase(db_path)

        # Initialize members in database
        self.db.initialize_members(self.cfg.assimilation.n_members)

        # Read experiment status from database
        self.load_status_from_db()

        self.paths = ExperimentPaths(experiment_path, self.cfg)
        self.obs = ExperimentObservations(self.cfg, self.cycles, self.paths)

    def load_status_from_db(self):
        """Load the status of the experiment from the database"""
        # Get experiment state
        self.current_cycle_i, self.filter_run, self.analysis_run = (
            self.db.get_experiment_state()
        )

        # Get member status
        members_data = self.db.get_all_members_status()
        self.members = []

        for member_i, advanced in members_data:
            # Get runtime statistics for this member
            runtime_stats = self.db.get_member_runtime_statistics(member_i)

            self.members.append(
                MemberStatus(
                    i=member_i, advanced=advanced, runtime_statistics=runtime_stats
                )
            )

        # Ensure we have the right number of members
        while len(self.members) < self.cfg.assimilation.n_members:
            member_i = len(self.members)
            self.members.append(
                MemberStatus(i=member_i, advanced=False, runtime_statistics=[])
            )

        self.members.sort(key=lambda m: m.i)

    def save_status_to_db(self):
        """Save the current status of the experiment to the database"""
        # Save experiment state
        self.db.set_experiment_state(
            self.current_cycle_i, self.filter_run, self.analysis_run
        )

        # Save member status
        for member in self.members:
            self.db.set_member_advanced(member.i, member.advanced)

    def set_next_cycle(self):
        """
        Update status to the next cycle
        """
        self.current_cycle_i += 1
        if self.current_cycle_i >= len(self.cycles):
            raise ValueError("No more cycles to run")
        self.filter_run = False
        self.analysis_run = False

        # Reset all members' advanced status in database
        self.db.reset_members_advanced()

        # Update local member status
        for member in self.members:
            member.advanced = False

        # Save to database
        self.save_status_to_db()

    def generate_perturbations(self, cycle_i: int):
        """
        Generates perturbations for a given cycle and stores them in `data/diag/perturbations`.
        """

        if len(self.cfg.perturbations.variables) == 0:
            logger.warning("No perturbations defined in config, skipping")
            return

        if self.cfg.perturbations.seed is not None:
            logger.warning(f"Setting random seed to {self.cfg.perturbations.seed}")
            np.random.seed(self.cfg.perturbations.seed)

        # Open the first wrfinput file to get the shape of each variable
        wrfinput = xr.open_dataset(self.paths.data_icbc / "wrfinput_d01_cycle_0")

        pert_file = self.paths.data_diag / "perturbations" / f"perts_cycle_{cycle_i}.nc"
        pert_file.parent.mkdir(parents=True, exist_ok=True)
        pert_file.unlink(missing_ok=True)

        # First, generate the perturbation field for each variable and member, and put them in a dataset
        perts = xr.Dataset()
        for var, pert_config in self.cfg.perturbations.variables.items():
            arr = np.stack(
                [
                    perturbations.generate_perturbation_field(
                        shape=wrfinput[var].shape,
                        mean=pert_config.mean,
                        sd=pert_config.sd,
                        rounds=pert_config.rounds,
                        boundary=pert_config.boundary,
                        min_value=pert_config.min_value,
                        max_value=pert_config.max_value,
                    )
                    for _ in range(self.cfg.assimilation.n_members)
                ],
                axis=0,
            )
            perts[var] = xr.DataArray(
                arr,
                dims=("member", *wrfinput[var].dims),
                coords={"member": range(self.cfg.assimilation.n_members)},
            )

        # Write the dataset to a netCDF file
        encoding = {
            var: {"zlib": True, "complevel": 5, "dtype": "float32"}
            for var in perts.data_vars
        }
        # Set chunksize for all variables: 1 on member and time, full size on rest
        for var in perts.data_vars:
            chunks = [1, 1] + [s for s in perts[var].shape[2:]]
            encoding[var]["chunksizes"] = chunks
        perts.attrs["experiment_name"] = self.cfg.metadata.name
        perts.attrs["cycle"] = cycle_i

        logger.info(f"Writing perturbations for {cycle_i} to {pert_file}")
        perts.to_netcdf(pert_file, encoding=encoding)

    def apply_perturbations(self, member_i: int):
        """
        Apply perturbations to the initial conditions of a member.
        Must be generated with `generate_perturbations` first.
        You should call this function once for every member, after the initial conditions
        are copied in the member directory (either during `ensemble setup` or `ensemble cycle`).
        """

        if member_i >= self.cfg.assimilation.n_members:
            raise ValueError(
                f"Member index {member_i} is out of bounds for {self.cfg.assimilation.n_members} members"
            )

        pert_file = (
            self.paths.data_diag
            / "perturbations"
            / f"perts_cycle_{self.current_cycle_i}.nc"
        )
        if not pert_file.exists():
            raise FileNotFoundError(
                f"Perturbation file for cycle {self.current_cycle_i} not found at {pert_file}"
            )

        perts = xr.open_dataset(pert_file).sel(member=member_i)

        with netCDF4.Dataset(
            self.paths.member_path(member_i) / "wrfinput_d01", "r+"
        ) as member_icbc:  # type: ignore
            for var, pert_config in self.cfg.perturbations.variables.items():
                logger.debug(f"Applying perturbation to {var} for member {member_i}")
                if var not in member_icbc.variables:
                    raise ValueError(f"Variable {var} not found in member IC/BC file.")

                field = perts[var].values
                if pert_config.operation == "add":
                    member_icbc[var][:] += field
                elif pert_config.operation == "multiply":
                    member_icbc[var][:] *= field
                else:
                    raise ValueError(
                        f"Unknown perturbation operation: {pert_config.operation}"
                    )

        logger.info(f"Applied perturbations to member {member_i}")

    def update_bc(self, member_i: int):
        """
        Run `update_wrf_bc` to update the boundary conditions of a member to match the initial conditions.
        """

        member_path = self.paths.member_path(member_i)
        icbc_target_file = member_path / "wrfinput_d01"
        bdy_target_file = member_path / "wrfbdy_d01"

        res = update_bc.update_wrf_bc(
            self.cfg,
            icbc_target_file,
            bdy_target_file,
            log_filename=f"update_bc_member_{member_i}.log",
        )
        if (
            res.returncode != 0
            or "update_wrf_bc Finished successfully" not in res.output
        ):
            logger.error(
                f"Member {member_i}: update_wrf_bc failed with exit code {res.returncode}"
            )
            logger.error(res.output)
            raise external.ExternalProcessFailed(res)

    def advance_member(self, member_idx: int, cores: int) -> bool:
        """
        Run WRF to advance a member to the next cycle.
        Initial and boundary condition files must already be present in the member directory.
        Will generate the appropriate namelist. Will move forecasts to the output directory.

        Args:
            member: Index of the member to advance
            cores: Number of cores to use

        Returns:
            True if the member was advanced successfully
        """

        member = self.members[member_idx]
        member_path = self.paths.member_path(member_idx)
        cycle = self.cycles[self.current_cycle_i]

        # Refuse to run model if already advanced
        if member.advanced:
            logger.error(f"Member {member_idx} already advanced")
            return False

        # Locate WRF executable, icbc, ensure they all exist
        wrf_exe_path = (member_path / "wrf.exe").resolve()
        if not wrf_exe_path.exists():
            logger.error(
                f"Member {member_idx}: WRF executable not found at {wrf_exe_path}"
            )
            return False

        ic_path = (member_path / "wrfinput_d01").resolve()
        bc_path = (member_path / "wrfbdy_d01").resolve()
        if not ic_path.exists() or not bc_path.exists():
            logger.error(
                f"Member {member_idx}: Initial/boundary conditions not found at {ic_path} or {bc_path}"
            )
            return False

        # Generate namelist
        wrf_namelist_path = member_path / "namelist.input"
        wrf.generate_wrf_namelist(
            self.cfg, cycle, True, wrf_namelist_path, member_idx, self.paths
        )

        # Clean old log files
        for f in member_path.glob("rsl.*"):
            f.unlink()

        # Run WRF
        logger.info(f"Running WRF for member {member_idx}...")
        cmd = [
            *self.cfg.slurm.mpirun_command.split(" "),
            "-n",
            str(cores),
            str(wrf_exe_path),
        ]

        start_time = dt.datetime.now()
        res = external.runc(cmd, cwd=member_path)
        end_time = dt.datetime.now()

        # Check output logs
        rsl_file = member_path / "rsl.out.0000"
        if not rsl_file.exists():
            logger.error(f"Member {member_idx}: RSL file not found at {rsl_file}")
            return False

        logger.add_log_file(rsl_file)
        rsl_content = rsl_file.read_text()

        if "SUCCESS COMPLETE WRF" not in rsl_content:
            logger.error(
                f"Member {member_idx}: wrf.exe failed with exit code {res.returncode}"
            )
            return False

        # Store logs in a zip file
        if logger.log_dir is not None:
            rsl_files = sorted(member_path.glob("rsl.*"))
            utils.zip_files(rsl_files, logger.log_dir / "rsl.zip")

        # Delete first output file
        first_output = (
            self.paths.scratch_forecasts_path(self.current_cycle_i, member_idx)
            / f"wrfout_d01_{cycle.start.strftime('%Y-%m-%d_%H:%M:%S')}"
        )
        if first_output.exists():
            logger.info(f"Removing first output file {first_output}")
            first_output.unlink()

        # Update member status using database instead of file locking
        # The database handles concurrency internally
        self.members[member_idx].advanced = True

        # Add runtime statistics to database
        self.db.add_runtime_statistics(
            member_idx,
            self.current_cycle_i,
            start_time,
            end_time,
            int((end_time - start_time).total_seconds()),
        )

        # Update member status in database
        self.db.set_member_advanced(member_idx, True)

        # Update local runtime statistics
        self.members[member_idx].runtime_statistics.append(
            RuntimeStatistics(
                cycle=self.current_cycle_i,
                start=start_time,
                end=end_time,
                duration_s=int((end_time - start_time).total_seconds()),
            )
        )

        return True

    def filter(self) -> bool:
        """
        Run the Kalman Filter for the current cycle

        Returns:
            True if the filter was run successfully
        """

        if self.filter_run:
            logger.error("Filter already run for current cycle")
            return False
        if not self.all_members_advanced:
            logger.error("Not all members have been advanced")
            return False

        dart_dir = self.cfg.directories.dart_root / "models" / "wrf" / "work"

        # Grab observations
        obs_seq = dart_dir / "obs_seq.out"
        obs_seq.unlink(missing_ok=True)

        obs_file = self.paths.obs / f"cycle_{self.current_cycle_i}.obs_seq"
        if not obs_file.exists():
            logger.error(f"Observation file for current cycle not found at {obs_file}")
            return False
        utils.copy(obs_file, obs_seq)

        # Write lists of input and output files
        # The input list is the latest forecast for each member
        wrfout_name = "wrfout_d01_" + self.current_cycle.end.strftime(
            "%Y-%m-%d_%H:%M:%S"
        )
        priors = [
            self.paths.scratch_forecasts_path(self.current_cycle_i, member_i)
            / wrfout_name
            for member_i in range(0, self.cfg.assimilation.n_members)
        ]
        posterior = [
            self.paths.scratch_dart_path(self.current_cycle_i)
            / f"dart_{prior.parent.name}.nc"
            for prior in priors
        ]

        dart_input_txt = dart_dir / "input_list.txt"
        dart_input_txt.write_text("\n".join(str(prior.resolve()) for prior in priors))
        logger.info(f"Wrote {dart_input_txt}")
        dart_output_txt = dart_dir / "output_list.txt"
        dart_output_txt.write_text("\n".join(str(post.resolve()) for post in posterior))
        logger.info(f"Wrote {dart_output_txt}")

        self.paths.scratch_dart_path(self.current_cycle_i).mkdir(exist_ok=True)

        # Link wrfinput, required by filter to read coordinates
        wrfinput_path = dart_dir / "wrfinput_d01"
        wrfinput_path.unlink(missing_ok=True)
        if self.cfg.data.per_member_meteorology:
            wrfinput_cur_cycle_path = (
                self.paths.data_icbc
                / "member_00"
                / f"wrfinput_d01_member_00_cycle_{self.current_cycle_i}"
            )
        else:
            wrfinput_cur_cycle_path = (
                self.paths.data_icbc / f"wrfinput_d01_cycle_{self.current_cycle_i}"
            )
        wrfinput_path.symlink_to(wrfinput_cur_cycle_path)
        logger.info(f"Linked {wrfinput_path} to {wrfinput_cur_cycle_path}")

        # Run filter
        if self.cfg.assimilation.filter_mpi_tasks == 1:
            logger.info("Running filter w/out MPI")
            cmd = ["./filter"]
        else:
            logger.info(
                f"Using MPI to run filter, n={self.cfg.assimilation.filter_mpi_tasks}"
            )
            cmd = [
                *self.cfg.slurm.mpirun_command.split(" "),
                "-n",
                str(self.cfg.assimilation.filter_mpi_tasks),
                "./filter",
            ]
        res = external.runc(cmd, dart_dir, log_filename="filter.log")
        if res.returncode != 0 or "Finished ... at" not in res.output:
            logger.error(f"filter failed with exit code {res.returncode}")
            return False

        # Keep obs_seq.final for diagnostics, convert to netcdf
        obs_seq_final = dart_dir / "obs_seq.final"
        utils.copy(
            obs_seq,
            self.paths.data_diag / f"cycle_{self.current_cycle_i}.obs_seq.final",
        )
        obs_seq_final_nc = self.paths.data_diag / f"cycle_{self.current_cycle_i}.nc"
        obs_sequence.obs_seq_to_nc(
            self.cfg.directories.dart_root, obs_seq_final, obs_seq_final_nc
        )

        self.filter_run = True
        self.save_status_to_db()
        return True

    def cycle_member(self, member_i: int, use_forecast: bool):
        """
        Merge the IC/BC of a member with the analysis from the previous cycle.
        Must be done for all members after finishing a set of forward runs and running
        the filter. If you have no observations and filter is skipped, run this step with
        use_forecast=True to use the forecast from the previous cycle as the analysis.

        Args:
            member_i: Index of the member to cycle
            use_forecast: Use the forecast from the previous cycle as the analysis
        """

        member_path = self.paths.member_path(member_i)
        next_cycle_i = self.current_cycle_i + 1

        # Find analysis/forecast file to use
        wrfout_name = "wrfout_d01_" + self.current_cycle.end.strftime(
            "%Y-%m-%d_%H:%M:%S"
        )
        if use_forecast:
            analysis_file = (
                self.paths.scratch_forecasts_path(self.current_cycle_i)
                / f"member_{member_i:02d}"
                / wrfout_name
            )
        else:
            analysis_file = (
                self.paths.scratch_analysis_path(self.current_cycle_i)
                / f"member_{member_i:02d}"
                / wrfout_name
            )

        if not analysis_file.exists():
            raise FileNotFoundError(analysis_file)
        logger.info(f"Using {analysis_file} as analysis for member {member_i}")

        # Copy the initial & boundary condition files for the next cycle, as is
        # First check if there is a member-specific IC/BC file for the next cycle,
        # otherwise use the ensemble default
        icbc_file = (
            self.paths.data_icbc
            / f"member_{member_i:02d}"
            / f"wrfinput_d01_member_{member_i:02d}_cycle_{next_cycle_i}"
        )
        if icbc_file.exists():
            logger.info(f"Using member-specific IC/BC file {icbc_file}")
        else:
            icbc_file = self.paths.data_icbc / f"wrfinput_d01_cycle_{next_cycle_i}"

        # First check if there is a member-specific BC file for the next cycle,
        # otherwise use the ensemble default
        bdy_file = (
            self.paths.data_icbc
            / f"member_{member_i:02d}"
            / f"wrfbdy_d01_member_{member_i:02d}_cycle_{next_cycle_i}"
        )
        if bdy_file.exists():
            logger.info(f"Using member-specific BC file {bdy_file}")
        else:
            bdy_file = self.paths.data_icbc / f"wrfbdy_d01_cycle_{next_cycle_i}"

        icbc_target_file = member_path / "wrfinput_d01"
        bdy_target_file = member_path / "wrfbdy_d01"

        utils.copy(icbc_file, icbc_target_file)
        utils.copy(bdy_file, bdy_target_file)

        # Copy cycled variables from the analysis file to the IC file
        with (
            netCDF4.Dataset(analysis_file, "r") as nc_analysis,  # type: ignore
            netCDF4.Dataset(icbc_target_file, "r+") as nc_icbc,  # type: ignore
        ):
            for name in self.cfg.assimilation.cycled_variables:
                if name not in nc_analysis.variables:
                    logger.warning(f"Member {member_i}: {name} not in analysis file")
                    continue
                logger.info(f"Member {member_i}: Copying {name}")
                nc_icbc[name][:] = nc_analysis[name][:]

            # Add experiment name to attributes
            nc_icbc.experiment_name = self.cfg.metadata.name

        # Remove forecast files, log files
        logger.info(f"Cleaning up member directory {member_path}")
        for f in member_path.glob("wrfout*"):
            logger.debug(f"Removing forecast file {f}")
            f.unlink()
        for f in member_path.glob("rsl*"):
            logger.debug(f"Removing log file {f}")
            f.unlink()

    def set_wrf_environment(self):
        """
        Adds the environment variables from config's `environment.wrf` to the current environment
        """

        for key, value in self.cfg.environment.wrf.items():
            os.environ[key] = value

    def set_dart_environment(self):
        """
        Adds the environment variables from config's `environment.dart` to the current environment
        """

        for key, value in self.cfg.environment.dart.items():
            os.environ[key] = value

    @property
    def current_cycle(self) -> cycling.CycleInformation:
        """
        Get the current cycle
        """

        return self.cycles[self.current_cycle_i]

    @property
    def all_members_advanced(self) -> bool:
        """
        Check if all ensemble members have been advanced
        """

        return all(m.advanced for m in self.members)



from dataclasses import dataclass
from pathlib import Path

import duckdb
import pandas as pd

from wrf_ensembly import external
from wrf_ensembly import observations as obs
from wrf_ensembly import wrf
from wrf_ensembly.config import Config
from wrf_ensembly.console import logger
from wrf_ensembly.cycling import CycleInformation
from wrf_ensembly.experiment.paths import ExperimentPaths


@dataclass
class ObservationFileMetadata:
    path: Path
    instrument: str
    start_time: pd.Timestamp
    end_time: pd.Timestamp


class ExperimentObservations:
    """Manages observation files for a WRF-Ensembly experiment."""

    def __init__(
        self, config: Config, cycles: list[CycleInformation], paths: ExperimentPaths
    ):
        self.cfg = config
        self.cycles = cycles
        self.paths = paths

    def _get_duckdb(self, read_only: bool) -> duckdb.DuckDBPyConnection:
        if read_only:
            return duckdb.connect(
                database=str(self.paths.obs_db),
                read_only=True,
            )

        # When also writing, ensure the observation table exists
        con = duckdb.connect(
            database=str(self.paths.obs_db),
            read_only=False,
        )
        con.execute("""
                    CREATE TABLE IF NOT EXISTS observations (
                        instrument STRING NOT NULL,
                        quantity STRING NOT NULL,
                        time TIMESTAMP NOT NULL,
                        longitude DOUBLE NOT NULL,
                        latitude DOUBLE NOT NULL,
                        x DOUBLE NOT NULL,
                        y DOUBLE NOT NULL,
                        z DOUBLE NOT NULL,
                        z_type STRING NOT NULL,
                        value DOUBLE,
                        value_uncertainty DOUBLE,
                        qc_flag INT NOT NULL,
                        orig_coords STRUCT(
                            indices INT[], shape INT[], names STRING[]
                        ) NOT NULL,
                        orig_filename STRING NOT NULL,
                        metadata STRING,
                        downsampling_info STRUCT(
                            method STRING,
                            obs_count INT,
                            time_spread_seconds DOUBLE,
                            spatial_spread_meters DOUBLE
                        )
            )""")
        return con

    def get_available_quantities(self) -> list[dict]:
        """Returns all combinations of instrument and quantity available in the database."""

        with self._get_duckdb(read_only=True) as con:
            result = con.execute("""
                SELECT instrument, quantity, COUNT(*) as count
                FROM observations
                GROUP BY instrument, quantity
                ORDER BY count DESC
            """).fetch_df()
        return result.to_dict(orient="records")

    def get_available_observations_overview(
        self,
    ) -> list[dict]:
        """Returns a dataframe with observation filenames and their row counts, min time and max time."""

        with self._get_duckdb(read_only=True) as con:
            result = con.execute("""
                SELECT
                    orig_filename as filename,
                    instrument as instrument,
                    MIN(time) as start_time,
                    MAX(time) as end_time,
                    COUNT(*) as count
                FROM observations
                GROUP BY orig_filename, instrument
                ORDER BY instrument, start_time
            """).fetch_df()
        return result.to_dict(orient="records")

    def delete_observation_file(self, filename: str) -> int:
        """
        Removes all observations from a specific file from the database.

        Args:
            filename: The name of the file to remove (not the full path)

        Returns:
            The number of observations removed.
        """

        with self._get_duckdb(read_only=False) as con:
            result = con.execute(
                f"DELETE FROM observations WHERE orig_filename = '{filename}'"
            )
            return result.rowcount

    def trim_observation_file(
        self, input_path: Path, output_path: Path
    ) -> tuple[str, int, int]:
        """
        Trims an observation file temporally and spatially according to the experiment configuration.
        The file is only created if there are observations left after trimming.

        This function assumes that preprocessing is completed (needs a wrfinput file to get the spatial bounds).

        Args:
            input_path: Path to the input observation file
            output_path: Path where the trimmed observation file will be saved

        Returns:
            The input file name, the number of observations in the original file, and the
            number of observations in the trimmed file.
        """

        filename = input_path.name
        df = obs.io.read_obs(input_path)
        original_len = len(df.index)

        # Find wrfinput file
        if not self.cfg.data.per_member_meteorology:
            wrfinput_path = self.paths.data_icbc / "wrfinput_d01_cycle_0"
        else:
            wrfinput_path = self.paths.data_icbc / "member_00" / "wrfinput_d01_cycle_0"
        if not wrfinput_path.exists():
            raise FileNotFoundError(
                f"wrfinput file not found at {wrfinput_path}, cannot trim observations spatially"
            )

        # Trim file into experiment time and space bounds
        transformer = wrf.get_wrf_proj_transformer(self.cfg.domain_control)
        start_time, end_time = wrf.get_temporal_domain_bounds(self.cycles)
        x_min, x_max, y_min, y_max = wrf.get_spatial_domain_bounds(wrfinput_path)

        df = obs.utils.project_locations_to_wrf(df, transformer)
        df = df[(df["time"] >= start_time) & (df["time"] <= end_time)]
        df["in_domain"] = (
            (df["x"] >= x_min)
            & (df["x"] <= x_max)
            & (df["y"] >= y_min)
            & (df["y"] <= y_max)
        )

        # For spatial filtering, we must make sure that the final array has no NaNs after
        # reshaping into the original shape.
        # Thus, we gotta check if after grouping by the original coordinates (keeping
        # the fastest-changing dimension out), there are any groups with at least one
        # observation inside the domain. Only the rest can be thrown out.
        # This proceedure must be done for each orig_filename,quantity pair separately
        per_quantity_dfs = []
        for _, df_subset in df.groupby(["orig_filename", "quantity"]):
            if df_subset.empty:
                continue

            # Find the fastest-changing dimension (the one with the smallest size in 'shape')
            orig_coords = df_subset["orig_coords"].iloc[0]
            shape = orig_coords["shape"]
            smalled_dim_index = shape.argmin()

            # Create a column with the groups, excluding the fastest-changing dimension
            df_subset["group_key"] = df_subset["orig_coords"].apply(
                lambda x: tuple(
                    idx for i, idx in enumerate(x["indices"]) if i != smalled_dim_index
                )
            )

            # Find groups with at least one observation inside the domain
            valid_groups = df_subset[df_subset["in_domain"]].groupby("group_key").size()
            valid_group_keys = valid_groups[valid_groups > 0].index

            # Keep only observations in valid groups
            df_subset = df_subset[df_subset["group_key"].isin(valid_group_keys)]

            # Set all outside-domain observations to NaN
            df_subset.loc[~df_subset["in_domain"], "value"] = pd.NA
            df_subset.loc[~df_subset["in_domain"], "value_uncertainty"] = pd.NA

            per_quantity_dfs.append(df_subset)

        if not per_quantity_dfs:
            trimmed_len = 0
        else:
            df = pd.concat(per_quantity_dfs, ignore_index=True)
            df = df.drop(columns=["in_domain", "group_key"])
            trimmed_len = len(df.index)

        # Save the trimmed observations to the output file
        if trimmed_len > 0:
            obs.io.write_obs(df, output_path)

        return filename, original_len, trimmed_len

    def add_observation_file(self, input_path: Path) -> int:
        """
        Adds an observation file to the experiment DuckDB database.

        This function assumes the file is already in the WRF-Ensembly observation format
        and has been trimmed as needed.

        Args:
            input_path: Path to the observation file to add to the database

        Returns:
            The number of observations added to the database.
        """

        df = obs.io.read_obs(input_path)
        if df.empty:
            return 0

        # Grab a connection to the database and save the observations
        with self._get_duckdb(read_only=False) as con:
            # First, if observations from this file already exist, remove them
            con.execute(
                f"DELETE FROM observations WHERE orig_filename = '{input_path.name}'"
            )

            con.register("df_view", df)
            con.execute("""
                INSERT INTO observations (
                    instrument, quantity, time, longitude, latitude, x, y, z, z_type,
                    value, value_uncertainty, qc_flag, orig_coords, orig_filename, metadata
                )
                SELECT
                    instrument, quantity, time, longitude, latitude, x, y, z, z_type,
                    value, value_uncertainty, qc_flag, orig_coords, orig_filename, metadata
                FROM df_view
            """)

        return len(df.index)

    def apply_superorbing(self) -> None:
        """
        Applies the configured downsampling (super-orbing) methods to all observations in the database.

        It will clear any old downsampled observations and replace them with the new ones.

        The `observations.superorbing` configuration dictionary controls how to downsample each instrument. The keys are the `instrument.quantity` pairs, e.g. `radiosonde.temperature`, and the values are how to downsample, check the `SuperorbingConfig` dataclass in `config.py` for details.
        """

        if not self.cfg.observations.superorbing:
            logger.info("No superorbing configuration found, skipping downsampling.")
            return

        with self._get_duckdb(read_only=False) as con:
            # Clean up any old downsampled observations
            res = con.execute(
                "DELETE FROM observations WHERE downsampling_info IS NOT NULL"
            )
            if res.rowcount > 0:
                logger.info(f"Removed {res.rowcount} old downsampled observations.")

            for key, superorb_config in self.cfg.observations.superorbing.items():
                try:
                    instrument, quantity = key.split(".")
                except ValueError:
                    raise ValueError(
                        f"Invalid superorbing key '{key}', must be in the format 'instrument.quantity'"
                    )

                logger.info(f"Applying superorbing for {instrument}.{quantity}'")

                df = con.execute(
                    f"SELECT * FROM observations WHERE instrument = '{instrument}' AND quantity = '{quantity}'"
                ).fetchdf()
                if df.empty:
                    logger.warning(
                        f"No observations found for {instrument}.{quantity}, skipping superorbing."
                    )
                    continue

                df_superobed = obs.superorbing.superorb_dbscan(df, superorb_config)

                # Insert the new superobed observations
                con.register("df_superobed_view", df_superobed)
                con.execute("""
                    INSERT INTO observations (
                        instrument, quantity, time, longitude, latitude, x, y, z, z_type,
                        value, value_uncertainty, qc_flag, orig_coords, orig_filename, metadata, downsampling_info
                    )
                    SELECT
                        instrument, quantity, time, longitude, latitude, x, y, z, z_type,
                        value, value_uncertainty, qc_flag, orig_coords, orig_filename, metadata, downsampling_info
                    FROM df_superobed_view
                """)

                logger.info(
                    f"Superorbing complete for {instrument}.{quantity}, reduced from {len(df)} to {len(df_superobed)} observations."
                )

    def get_observations_for_cycle(
        self, cycle: CycleInformation
    ) -> pd.DataFrame | None:
        """
        Retrieves observation data for a specific cycle and set of instruments.

        Args:
            cycle: The cycle information to filter observations. The assimilation window from `cfg.assimilation.half_window_length_minutes` is applied.
        """

        instruments = self.cfg.observations.instruments_to_assimilate

        start_time = cycle.start - pd.Timedelta(
            minutes=self.cfg.assimilation.half_window_length_minutes
        )
        end_time = cycle.end + pd.Timedelta(
            minutes=self.cfg.assimilation.half_window_length_minutes
        )

        # Query the observations with duck db, find only files that overlap with the time window and instrument list
        with self._get_duckdb(read_only=True) as con:
            # If there are no super-orbed observations, we will use the original ones.
            superorbed_count = con.execute("""
                SELECT COUNT(*) FROM observations WHERE downsampling_info IS NOT NULL
            """).fetchone()
            superorbed_count = superorbed_count[0] if superorbed_count else 0
            superorbed_available = superorbed_count > 0

            if superorbed_available:
                logger.info("Using super-orbed observations for assimilation.")
                observations = observations = con.execute(
                    f"SELECT * FROM observations WHERE time >= '{start_time}' AND time <= '{end_time}' WHERE downsampling_info IS NOT NULL"
                ).fetchdf()
            else:
                logger.info(
                    "No super-orbed observations found, using original observations."
                )
                observations = con.execute(
                    f"SELECT * FROM observations WHERE time >= '{start_time}' AND time <= '{end_time}' AND downsampling_info IS NULL"
                ).fetchdf()

        if instruments is not None:
            observations = observations[observations["instrument"].isin(instruments)]

        return observations if not observations.empty else None

    def convert_cycle_to_dart(self, cycle: CycleInformation):
        """Converts the observations for a given cycle to DART obs_seq format."""

        df = self.get_observations_for_cycle(cycle)
        if df is None or df.empty:
            logger.info(
                f"No observations for cycle {cycle.index}, skipping DART conversion"
            )
            return

        output_path = self.paths.obs / f"obs_seq.{cycle.index:03d}"
        dart_process = obs.dart.convert_to_dart_obs_seq(
            dart_path=self.cfg.directories.dart_root,
            observations=df,
            output_location=output_path,
        )
        logger.info(
            f"Converting observations for cycle {cycle.index} to DART obs_seq..."
        )
        result = external.run(dart_process)
        if result.returncode != 0:
            logger.error(result.output)
            raise RuntimeError(
                f"DART conversion failed for cycle {cycle.index} with return code {result.returncode}"
            )
        logger.info(f"Wrote DART obs_seq file to {output_path}")



"""Some utility commands for performing basic operations on observation files, such as dump, info, etc."""

import json
from pathlib import Path

import click
import pandas as pd
from rich.progress import track
from rich.table import Table

from wrf_ensembly import external
from wrf_ensembly.console import console
from wrf_ensembly.observations import dart as observations_dart
from wrf_ensembly.observations import io as obs_io


@click.command()
@click.argument("file_list", nargs=-1, type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(path_type=Path))
def join_files(file_list: list[Path], output_path: Path):
    """Given a list of WRF-ensembly observation files, join them into a single file."""

    dfs = []
    for file_path in track(file_list, description="Reading files..."):
        df = obs_io.read_obs(file_path)
        dfs.append(df)

    combined_df = pd.concat(dfs, ignore_index=True)
    obs_io.write_obs(combined_df, output_path)
    print(
        f"Successfully joined {len(file_list)} files into {output_path}, total observations: {len(combined_df)}"
    )


@click.command()
@click.argument("file_path", type=click.Path(exists=True, path_type=Path))
@click.option("--as-json", is_flag=True, help="Output in JSON format", default=False)
def dump_info(file_path: Path, as_json: bool):
    """
    Print basic information about a WRF-Ensembly observation file.
    - Number of observations
    - Instruments present and their counts
    - z_types present and their counts
    - Quantities present and their counts
    - Time range of observations
    - Geographical range (min/max latitude and longitude)
    """

    df = obs_io.read_obs(file_path)
    info = {
        "file_path": str(file_path),
        "num_observations": len(df),
        "instruments": df["instrument"].value_counts().to_dict(),
        "z_types": df["z_type"].value_counts().to_dict(),
        "quantities": df["quantity"].value_counts().to_dict(),
        "time_range": {
            "start": df["time"].min().isoformat(),
            "end": df["time"].max().isoformat(),
        },
        "geographical_range": {
            "latitude": {
                "min": df["latitude"].min(),
                "max": df["latitude"].max(),
            },
            "longitude": {
                "min": df["longitude"].min(),
                "max": df["longitude"].max(),
            },
        },
    }

    if as_json:
        print(json.dumps(info, indent=2))
        return 0

    # Pretty print using rich
    console.print(f"[bold]Observation File Info: {file_path}[/bold]")

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Property", style="dim")
    table.add_column("Value")
    table.add_row("Number of Observations", str(info["num_observations"]))
    table.add_row(
        "Instruments", ", ".join(f"{k} ({v})" for k, v in info["instruments"].items())
    )
    table.add_row(
        "Z Types", ", ".join(f"{k} ({v})" for k, v in info["z_types"].items())
    )
    table.add_row(
        "Quantities", ", ".join(f"{k} ({v})" for k, v in info["quantities"].items())
    )
    table.add_row(
        "Time Range", f"{info['time_range']['start']} to {info['time_range']['end']}"
    )
    table.add_row(
        "Latitude Range",
        f"{info['geographical_range']['latitude']['min']} to {info['geographical_range']['latitude']['max']}",
    )
    table.add_row(
        "Longitude Range",
        f"{info['geographical_range']['longitude']['min']} to {info['geographical_range']['longitude']['max']}",
    )
    console.print(table)


@click.command()
@click.argument("file_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(path_type=Path))
@click.option("--min-lat", type=float, required=False, help="Minimum latitude")
@click.option("--max-lat", type=float, required=False, help="Maximum latitude")
@click.option("--min-lon", type=float, required=False, help="Minimum longitude")
@click.option("--max-lon", type=float, required=False, help="Maximum longitude")
@click.option(
    "--min-time", type=str, required=False, help="Minimum timestamp (ISO format)"
)
@click.option(
    "--max-time", type=str, required=False, help="Maximum timestamp (ISO format)"
)
def filter_obs(
    file_path: Path,
    output_path: Path,
    min_lat: float | None,
    max_lat: float | None,
    min_lon: float | None,
    max_lon: float | None,
    min_time: str | None,
    max_time: str | None,
):
    """Filters an observation file based on spatial and temporal criteria."""

    df = obs_io.read_obs(file_path)
    original_len = len(df)

    if min_lat is not None:
        if min_lat < -90 or min_lat > 90:
            raise ValueError("min_lat must be between -90 and 90")
        df = df[df["latitude"] >= min_lat]
    if max_lat is not None:
        if max_lat < -90 or max_lat > 90:
            raise ValueError("max_lat must be between -90 and 90")
        df = df[df["latitude"] <= max_lat]
    if min_lon is not None:
        if min_lon < -180 or min_lon > 180:
            raise ValueError("min_lon must be between -180 and 180")
        df = df[df["longitude"] >= min_lon]
    if max_lon is not None:
        if max_lon < -180 or max_lon > 180:
            raise ValueError("max_lon must be between -180 and 180")
        df = df[df["longitude"] <= max_lon]
    if min_time is not None:
        try:
            min_time_parsed = pd.to_datetime(min_time)
        except Exception as e:
            raise ValueError(f"Invalid min_time format: {e}")
        df = df[df["time"] >= min_time_parsed]
    if max_time is not None:
        try:
            max_time_parsed = pd.to_datetime(max_time)
        except Exception as e:
            raise ValueError(f"Invalid max_time format: {e}")
        df = df[df["time"] <= max_time_parsed]

    print(
        f"Filtered observations from {original_len} to {len(df)} based on spatial criteria."
    )
    obs_io.write_obs(df, output_path)


@click.command()
@click.option(
    "--dart-path", type=click.Path(exists=True, path_type=Path), required=True
)
@click.argument("obs_file", type=click.Path(exists=True, path_type=Path))
@click.argument("output_file", type=click.Path(path_type=Path))
def to_obs_seq(obs_file: Path, output_file: Path, dart_path: Path):
    """
    Convert a WRF-Ensembly observation file to DART obs_seq format.

    You must have built the `wrf_ensembly` observation converter in DART for this to work, check the `DART/observations/obs_converters/wrf_ensembly` directory.

    Args:
        obs_file: Path to the input WRF-Ensembly observation file.
        output_file: Path where the output DART obs_seq file will be saved.
        dart_path: Path to the root of the DART installation.

    Example usage:
        wrf-ensembly-obs convert-to-obs-seq --dart-path /path/to/DART /path/to/input.parquet /path/to/output.obs_seq
    """

    process = observations_dart.convert_to_dart_obs_seq(
        dart_path=dart_path,
        observations=obs_io.read_obs(obs_file),
        output_location=output_file,
    )
    result = external.run(process)
    print("Converter output:")
    print(result.output)

    if result.returncode != 0:
        print("Error converting to DART obs_seq!")
    else:
        print(f"Successfully converted to DART obs_seq: {output_file}")



"""
Handles interactions with the "update_wrf_bc" executable that updates the boundary conditions
to match the initial conditions, after they are modified by cycling or applying perturbations.
"""

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

from wrf_ensembly import utils
from wrf_ensembly.config import Config
from wrf_ensembly.external import ExternalProcess, run


def update_wrf_bc(
    cfg: Config,
    wrfinput: Path,
    wrfbdy: Path,
    log_filename: Optional[str] = None,
):
    """
    Updates the given `wrfbdy` file to match the `wrfinput` file, using `update_wrf_bc`.
    Required if you have modified the `wrfinput` file.

    The `update_wrf_bc` executable is expected to be found inside the DART work directory.
    It will be linked in a temp. directory so that you can run this function in parallel.

    Args:
        cfg: The configuration object.
        wrfinput: The wrfinput file to update from.
        wrfbdy: The wrfbdy file to update. Will be mutated.
        log_dir: The directory to log the stdout and stderr of the process to.
        log_filename: The filename inside `log_dir` to log the stdout and stderr

    Returns:
        The result of the external process call (see `ExternalProcessResult`).
    """

    with TemporaryDirectory(prefix="wrf_ensembly_update_bc") as work_dir:
        work_dir = Path(work_dir)

        # Link input files
        if not wrfinput.exists():
            raise FileNotFoundError(f"wrfinput file not found: {wrfinput}")
        if not wrfbdy.exists():
            raise FileNotFoundError(f"wrfbdy file not found: {wrfbdy}")

        sym_wrfinput = work_dir / "wrfinput_d01"
        sym_wrfbdy = work_dir / "wrfbdy_d01"
        sym_wrfinput.symlink_to(wrfinput.resolve())
        sym_wrfbdy.symlink_to(wrfbdy.resolve())

        # Link executable
        bin_path = (
            cfg.directories.dart_root / "models" / "wrf" / "work" / "update_wrf_bc"
        )
        if not bin_path.is_file():
            raise FileNotFoundError(f"update_wrf_bc executable not found in {bin_path}")

        command = work_dir / "update_wrf_bc"
        command.symlink_to(bin_path.resolve())

        # Copy namelist
        utils.copy(
            cfg.directories.dart_root / "models" / "wrf" / "work" / "input.nml",
            work_dir / "input.nml",
        )

        res = run(
            ExternalProcess(
                [str(command.resolve())],
                cwd=work_dir,
                log_filename=log_filename,
            )
        )

    return res



import datetime as dt
import re
import shutil
import tempfile
from dataclasses import dataclass, field
from pathlib import Path

from mashumaro import field_options
from mashumaro.mixins.toml import DataClassTOMLMixin

from wrf_ensembly import external, fortran_namelists, utils
from wrf_ensembly.config import Config
from wrf_ensembly.console import logger


@dataclass
class Observation:
    """Represents one observation file"""

    start_date: dt.datetime = field(
        metadata=field_options(
            deserialize=lambda v: dt.datetime.strptime(v, "%Y%m%dT%H%M%S").replace(
                tzinfo=dt.timezone.utc
            )
        )
    )
    """Timestamp of the first datapoint inside the file"""

    end_date: dt.datetime = field(
        metadata=field_options(
            deserialize=lambda v: dt.datetime.strptime(v, "%Y%m%dT%H%M%S").replace(
                tzinfo=dt.timezone.utc
            )
        )
    )
    """Timestamp of the last datapoint inside the file"""

    path: Path
    """Path to the physical file"""


@dataclass
class ObservationGroup(DataClassTOMLMixin):
    """
    Represents one group of observations, in which all files share the same format, use
    the same observation operator and are of the same DART kind
    """

    kind: str
    """DART kind for the observation"""

    converter: str
    """
    Path to converter executable. Must take input and output file as first and second
    arguments. You can add any other arguments you want to pass to the converter in the
    TOML file, they will be passed to the converter as well, but before the input and
    output file.
    """

    files: list[Observation]
    """Files in this group"""

    cwd: Path = Path(".")
    """Working directory for the converter"""

    def get_files_in_window(
        self, start: dt.datetime, end: dt.datetime
    ) -> list[Observation]:
        """Returns all files that overlap with the given window"""

        return [f for f in self.files if f.start_date <= end and start <= f.end_date]

    def convert_file(self, file: Observation, output_file: Path):
        """Converts a file to DART obs_seq format by running the appropriate converter"""

        res = external.run(
            external.ExternalProcess(
                [*self.converter.split(" "), file.path, output_file],
                cwd=Path(self.converter).parent,
                log_filename=f"converter_{file.path.name}.log",
            )
        )
        if res.returncode != 0:
            raise external.ExternalProcessFailed(res)
        if not output_file.exists():
            logger.error("Converter did not produce an output file!")
            raise external.ExternalProcessFailed(res)
        return res


def read_observation_group(path: Path) -> ObservationGroup:
    """Read an observation group file from the disk"""

    return ObservationGroup.from_toml(path.read_text())


def read_observations(dir: Path) -> dict[str, ObservationGroup]:
    """Reads all observation group files in a directory"""
    return {f.stem: read_observation_group(f) for f in dir.glob("*.toml")}


def join_obs_seq(
    cfg: Config,
    obs_seq_files: list[Path],
    output_file: Path,
    obs_kinds: list[str],
    binary_obs_sequence: bool = False,
):
    """
    Runs the `obs_sequence_tool` to join a list of obs_seq files into one.
    The tool must be compiled w/ DART and be available in the model's work directory.

    TODO Add temporal and spatial cropping options

    Args:
        cfg: Experiment configuration object
        obs_seq_files: List of obs_seq files to join
        output_file: Where to write the output file
        obs_kinds: List of observation kinds to include in the output file
        binary_obs_sequence: Whether to write the output file as a binary obs_seq file, defaults to False (i.e., ASCII)
    """

    dart_work_dir = cfg.directories.dart_root / "models" / "wrf" / "work"
    obs_seq_tool = dart_work_dir / "obs_sequence_tool"

    nml = {
        "obs_sequence_tool_nml": {
            "filename_seq_list": "./input_list",
            "filename_out": output_file.resolve(),
            "gregorian_cal": True,
        },
        "obs_sequence_nml": {"write_binary_obs_sequence": binary_obs_sequence},
        "obs_kind_nml": {
            "assimilate_these_obs_types": obs_kinds,
        },
        "location_nml": {},
        "utilities_nml": {
            "TERMLEVEL": 1,
            "module_details": False,
            "logfilename": "obs_sequence_tool.out",
            "nmlfilename": "obs_sequence_tool.nml",
            "write_nml": "file",
        },
    }

    # Link obs_sequence_tool inside a temp directory, create namelist
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        obs_seq_tool_ln = tmp_dir / "obs_sequence_tool"
        obs_seq_tool_ln.symlink_to(obs_seq_tool)

        # Write the namelist file
        namelist_path = tmp_dir / "input.nml"
        fortran_namelists.write_namelist(nml, namelist_path)

        # Write all input files inside a text file
        filelist_path = tmp_dir / "input_list"
        filelist_path.write_text("\n".join(str(f.resolve()) for f in obs_seq_files))
        print(filelist_path.read_text())

        # Call obs_sequence_tool, check results
        res = external.run(
            external.ExternalProcess(
                [obs_seq_tool_ln], cwd=tmp_dir, log_filename="obs_sequence_tool.log"
            )
        )
        if res.returncode != 0:
            logger.error(f"obs_sequence_tool exited with error code {res.returncode}!")
            logger.error(res.output)
            raise RuntimeError(f"obs_sequence_tool failed with code {res.returncode}")

    return res


def obs_seq_to_nc(
    dart_root: Path,
    obs_seq: Path,
    nc: Path,
    binary_obs_sequence: bool = False,
) -> external.ExternalProcessResult:
    """
    Uses the `obs_seq_to_netcdf` program to convert the given obs_seq file to netcdf format

    Args:
        dart_root: Path to the DART root directory
        obs_seq: Path to obs_seq file
        nc: Path to output netcdf file
        binary_obs_sequence: Whether the obs_seq file is binary or not, defaults to False

    Returns:
        The result of the external process
    """

    # Locate executable
    binary = dart_root / "models" / "wrf" / "work" / "obs_seq_to_netcdf"
    binary = binary.resolve()
    if not binary.exists():
        raise RuntimeError(f"obs_seq_to_netcdf binary not found at {binary}")

    # Link obs_seq_to_netcdf inside a temp directory
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)

        binary_ln = tmp_dir / "obs_seq_to_netcdf"
        binary_ln.symlink_to(binary)

        # Create namelist file
        nml = {
            "obs_seq_to_netcdf_nml": {
                "obs_sequence_name": str(obs_seq.resolve()),
                "obs_sequence_list": "",
                "append_to_netcdf": False,
                "lonlim1": 0.0,
                "lonlim2": 360,
                "latlim1": -90,
                "latlim2": 90,
                "verbose": False,
            },
            "obs_sequence_nml": {"write_binary_obs_sequence": binary_obs_sequence},
            "location_nml": {},
            "obs_kind_nml": {},
            "schedule_nml": {},
            "utilities_nml": {
                "TERMLEVEL": 1,
                "module_details": False,
                "logfilename": "obs_sequence_tool.out",
                "nmlfilename": "obs_sequence_tool.nml",
                "write_nml": "file",
            },
        }
        fortran_namelists.write_namelist(nml, tmp_dir / "input.nml")

        # Call obs_seq_to_netcdf
        res = external.run(external.ExternalProcess([binary_ln], cwd=tmp_dir))

        if res.returncode != 0:
            logger.error(f"obs_seq_to_netcdf exited with error code {res.returncode}!")
            logger.error(res.output)
            raise RuntimeError(f"obs_seq_to_netcdf failed with code {res.returncode}")

        # Move output file to the desired location
        obs_seq_out = tmp_dir / "obs_epoch_001.nc"
        utils.copy(obs_seq_out, nc)
        logger.info(f"Converted obs_seq file to netcdf: {nc}")

    return res


def is_obs_seq_empty(file: Path) -> bool:
    """
    Returns true if an obs_seq file is empty. To do this, we read the first couple of
    lines to get the num_obs value. The file is parsed with regex, thus this works only
    with ASCII obs_seq files.

    TODO: Do this in FORTRAN maybe?
    """

    # Read the first couple of lines (2KB)
    with file.open("r") as f:
        content = f.read(2048)

    # Use regex to find num_obs
    match = re.search(r"[^_]num_obs:\s+(\d+)", content)
    if match:
        num_obs = int(match.group(1))
        return num_obs == 0

    # If no match is found, assume the file is empty
    return True


def preprocess_for_wrf(
    dart_root: Path,
    wrfinput: Path,
    obs_seq: Path,
    assimilation_window_dart_time: tuple[int, int],
    boundary_width=0.0,
    boundary_error_factor=0.0,
    boundary_error_width=0.0,
    binary_obs_sequence=False,
):
    """
    Runs a obs_seq file through the `wrf_dart_obs_preprocess` utility that (supported features):
    - Removes all observations outside the domain
    - Inflate errors near the boundaries

    Args:
        dart_root: Path to DART root directory
        wrfinput: Path to the WRF input file
        obs_seq: Path to the obs_seq file. Will be overwritten with the preprocessed version, if successful.
        assimilation_window_dart_time: Tuple with the start and end of the assimilation window in DART time units (days since 1601-01-01, seconds since midnight)
        boundary_width: If > 0, the domain will be reduced by this amount of gridpoints.
        boundary_error_factor: If >0, the errors near the boundary will be increased by this factor.
        boundary_error_width: How near the boundary to increase the errors, in gridpoints. A ramping is applied, so the innermost points are changed by 0 and the outermost by `boundary_error_factor`.
        binary_obs_sequence: Whether the obs_seq file is binary or not, defaults to False
    """

    # Locate the executable
    binary = dart_root / "models" / "wrf" / "work" / "wrf_dart_obs_preprocess"
    binary = binary.resolve()
    if not binary.exists():
        raise RuntimeError(f"wrf_dart_obs_preprocess binary not found at {binary}")

    # Link wrf_dart_obs_preprocess inside a temp directory, alongside all input files
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)

        binary_ln = tmp_dir / "wrf_dart_obs_preprocess"
        binary_ln.symlink_to(binary)

        # Link input files
        wrfinput_ln = tmp_dir / "wrfinput_d01"
        wrfinput_ln.symlink_to(wrfinput)

        obs_seq_ln = tmp_dir / "obs_seq.in"
        obs_seq_ln.symlink_to(obs_seq)

        # Create namelist file
        nml = {
            "wrf_obs_preproc_nml": {
                # I/O
                "file_name_input": "obs_seq.in",
                "file_name_output": "obs_seq.out",
                # Time management, not supported
                "overwrite_obs_time": False,
                # Boundary
                "obs_boundary": boundary_width,
                "increase_bdy_error": boundary_error_factor > 0,
                "maxobsfac": boundary_error_factor,
                "obsdistbdy": boundary_error_width,
                # Surface elevation, not used
                "sfc_elevation_check": False,
                "sfc_elevation_tol": 3000.0,
                "obs_pressure_top": 0.0,
                "obs_height_top": 2.0e10,
                # Radiosonde stuff, not used
                "include_sig_data": True,
                "tc_sonde_radii": -1.0,
                # Superorbing, not used
                "superob_aircraft": False,
                "aircraft_horiz_int": 800.0,
                "aircraft_pres_int": 25000.0,
                "superob_sat_winds": False,
                "sat_wind_horiz_int": 800.0,
                "sat_wind_pres_int": 25000.0,
                # Overwrite QC flags, not used
                "overwrite_ncep_satwnd_qc": False,
                "overwrite_ncep_sfc_qc": False,
            },
            "obs_sequence_nml": {"write_binary_obs_sequence": binary_obs_sequence},
            "location_nml": {},
            "obs_kind_nml": {},
            "schedule_nml": {},
            "model_nml": {},
            "ensemble_manager_nml": {},
            "utilities_nml": {
                "TERMLEVEL": 1,
                "module_details": False,
                "logfilename": "obs_sequence_tool.out",
                "nmlfilename": "obs_sequence_tool.nml",
                "write_nml": "file",
            },
        }
        fortran_namelists.write_namelist(nml, tmp_dir / "input.nml")

        # Call utility, move result back to the original location
        stdin = (
            f"{assimilation_window_dart_time[0]} {assimilation_window_dart_time[1]}\n"
        )
        res = external.run(
            external.ExternalProcess([binary_ln], cwd=tmp_dir, stdin=stdin)
        )

        if res.returncode != 0:
            logger.error(
                f"wrf_dart_obs_preprocess exited with error code {res.returncode}!"
            )
            logger.error(res.output)
            raise external.ExternalProcessFailed(res)

        # If the file is empty, just remove the original
        obs_seq_out = tmp_dir / "obs_seq.out"
        if is_obs_seq_empty(obs_seq_out):
            logger.warning("Preprocessed obs_seq file is empty (all obs removed)!")
            obs_seq.unlink()
            return

        # Move output file to the desired location
        logger.info(f"Removing {obs_seq}")
        obs_seq.unlink()

        logger.info(f"Renaming {obs_seq_out} to {obs_seq}")
        shutil.move(obs_seq_out, obs_seq)



"""Main CLI for WRF-Ensembly observation converters.

This module provides a CLI for converting various observation data formats
to the standardized WRF-Ensembly observation format.

Adding a new converter:
1. Create a converter function in wrf_ensembly/observation/converters/your_format.py
   - The function should take input parameters and return a pandas DataFrame
   - The DataFrame must conform to the WRF-Ensembly observation schema
   - Use validate_schema() from wrf_ensembly.observation.io to check compliance

2. Add a CLI command in the same file using click decorators:
   - Use click decorators to define arguments and options
   - Call your converter function and save results using obs_io.write_obs()
   - See the aeronet.py file for an example structure

3. Import and add your command to this file:
   - from wrf_ensembly.observation.converters.your_format import your_format_cli
   - cli.add_command(your_format_cli)

4. Update the __init__.py files to include your new modules in __all__

The resulting CLI will be: wrf-ensembly-obs-convert your_format [args...]
"""

import click

from wrf_ensembly.observations.converters import aeronet_cli, remotap_spexone_cli
from wrf_ensembly.observations.operations import (
    dump_info,
    filter_obs,
    join_files,
    to_obs_seq,
)


@click.group()
def cli():
    pass


@cli.group()
def convert_group():
    """Convert raw observation files to WRF-Ensembly observation format.

    This tool provides converters for various observation data formats.
    Each converter takes raw observation files and converts them to the
    standardized WRF-Ensembly observation format (parquet files).
    """
    pass


convert_group.add_command(aeronet_cli)
convert_group.add_command(remotap_spexone_cli)


@cli.group()
def operations_group():
    """Perform operations on WRF-Ensembly observation files.

    This tool provides various operations for WRF-Ensembly observation files,
    such as joining multiple files, filtering observations, and dumping file info.
    """
    pass


operations_group.add_command(join_files)
operations_group.add_command(dump_info)
operations_group.add_command(filter_obs)
operations_group.add_command(to_obs_seq)

if __name__ == "__main__":
    cli()
